(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_Complexity
imports
  Reduction_Pair_Impl
  Reduction_Pair_Implementations
  Matchbounds_Impl
begin

subsection {*Proving complexity of TRSs*}

datatype_new ('f, 'v) complexity_proof =
  Rule_Shift_Complexity "'f redtriple_impl" "('f, 'v) rules" "('f, 'v) complexity_proof" |
  RisEmpty_Complexity |
  Remove_Nonapplicable_Rules_Complexity "('f,'v)rules" "('f,'v) complexity_proof" |
  Matchbounds_Complexity "('f,'v)bounds_info" |
  Matchbounds_Rel_Complexity "('f,'v)bounds_info" "('f,'v)rules" "('f,'v) complexity_proof"


primrec
  check_complexity_proof ::
    "shows \<Rightarrow>
     ('tp, 'f::{key,show}, 'v :: {key,show}) tp_ops \<Rightarrow> 'tp \<Rightarrow>
     ('f,'v)complexity_measure \<Rightarrow> complexity_class \<Rightarrow>
     ('f, 'v) complexity_proof \<Rightarrow>
     shows check" 
where
  "check_complexity_proof i I tp cm cc (Rule_Shift_Complexity redp Rdelete prf) = debug (i []) ''Rule Removal Complexity'' (do {
         tp' \<leftarrow> rule_shift_complexity_tt I (get_redtriple redp) Rdelete cm cc tp
          <+? (\<lambda>s. i +@+ '': error when applying the rule shifting technique on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_complexity_proof (i +@+ shows_string ''.1'') I tp' cm cc prf
           <+? (\<lambda>s. i +@+ '': error below the rule shifting processor'' +#+ shows_nl
             +@+ indent s)}
)" 
 | "check_complexity_proof i I tp cm cc (RisEmpty_Complexity) = debug (i []) ''R is empty for complexity'' (do {
     check (tp_ops.R I tp = []) (i +@+ '': R is not empty in '' +#+ shows_nl +@+ shows_tp I tp) })"
 | "check_complexity_proof i I tp cm cc (Remove_Nonapplicable_Rules_Complexity r prf) = debug (i []) ''Removing non-applicable rules'' (do {
       let R = tp_ops.R I tp;
       check_non_applicable_rules (tp_ops.is_QNF I tp) r
         <+? (\<lambda> s. i +@+ '': error when removing non-applicable rules'' +#+ shows_nl +@+ indent (shows_rule s +@+ shows '' is applicable''));
       let tp' = tp_spec.delete_rules I tp r;
       check_complexity_proof (i +@+ shows_string ''.1'') I tp' cm cc prf
         <+? (\<lambda>s. i +@+ '': error below the non-applicable rules removal'' +#+ shows_nl
           +@+ indent s)})"
 | "check_complexity_proof i I tp cm cc (Matchbounds_Complexity info) = debug (i []) ''Matchbounds'' (
       bounds_complexity I info cm cc tp 
         <+? (\<lambda>s. i +@+ '': error when applying matchbounds'' +#+ shows_nl +@+ indent s))"
 | "check_complexity_proof i I tp cm cc (Matchbounds_Rel_Complexity info Rdel prf) = debug (i []) ''Matchbounds-Rel'' (
      do {
        tp' <- bounds_complexity_rel I info Rdel cm cc tp 
           <+? (\<lambda>s. i +@+ '': error when applying relative matchbounds'' +#+ shows_nl +@+ indent s);
        check_complexity_proof (i +@+ shows_string ''.1'') I tp' cm cc prf
         <+? (\<lambda>s. i +@+ '': error below applying relative matchbounds'' +#+ shows_nl
           +@+ indent s)})"

lemma check_complexity_proof_sound: 
  assumes I: "tp_spec I" 
  and inf: "infinite (UNIV :: 'f set)"
  and ok: "isOK (check_complexity_proof i I tp cm cc prf)"
  shows "deriv_bound_measure_class (rel_qrstep (tp_ops.qreltrs I tp)) (cm :: ('f :: {key,show},'v :: {key,show})complexity_measure) cc"
proof -
  interpret tp_spec I by fact
  from ok show ?thesis
  proof (induct "prf" arbitrary: tp i cm cc)
    case RisEmpty_Complexity
    thus ?case by simp
  next
    case (Rule_Shift_Complexity rp Rdelete prof)
    note IH = Rule_Shift_Complexity 
    note res = IH(2)[simplified]    
    from res obtain tp' where rs: "rule_shift_complexity_tt I (get_redtriple rp) Rdelete cm cc tp = return tp'" by auto
    from res[unfolded rs, simplified] have rec: "isOK(check_complexity_proof (i +@+ shows_string ''.1'') I tp' cm cc prof)" by auto
    from IH(1)[OF this] have IH: "deriv_bound_measure_class (rel_qrstep (qreltrs tp')) cm cc" .
    from rule_shift_complexity_tt[OF I get_redtriple, OF rs IH] show ?case .
  next
    case (Remove_Nonapplicable_Rules_Complexity Rdelete prof)
    note IH = Remove_Nonapplicable_Rules_Complexity
    note res = IH(2)[simplified]
    from res have check: "isOK (check_non_applicable_rules (\<lambda>t. t \<in> NF_terms (set (Q tp))) Rdelete)" 
      and rec: " isOK (check_complexity_proof (i +@+ shows_string ''.1'') I (delete_rules tp Rdelete) cm cc prof)" by auto
    note IH = IH(1)[OF rec]
    from check_non_applicable_rules[OF refl check] have id: "\<And> R. qrstep (NFS tp) (set (Q tp)) (R - set Rdelete) = qrstep (NFS tp) (set (Q tp)) R" .
    from IH show ?case by (simp add: id)
  next
    case (Matchbounds_Complexity info) note IH = this
    show ?case 
      by (rule bounds_complexity[OF I], insert IH inf, auto)
  next
    case (Matchbounds_Rel_Complexity info del) note IH = this
    show ?case 
      by (rule bounds_complexity_rel[OF I], insert IH inf, auto)
  qed
qed

end
