(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Code_Setup_Haskell
imports
  Util
begin

subsection {*Code-Generator Setup*}

text {*The order of the final two arguments of @{const foldr} in
Isabelle is different from Haskell's @{text foldr}*}
definition flipped_foldr where "flipped_foldr f \<equiv> flip (foldr f)"
text {*Separate infix use and prefix use of infix operators*}
definition comp_infix where "comp_infix x y \<equiv> x \<circ> y"
definition Cons_infix where "Cons_infix x y \<equiv> x # y"

lemma foldr_flipped_foldr[code_unfold]:
  "foldr f xs a = flipped_foldr f a xs"
  by (simp add: flipped_foldr_def)

lemma flip_foldr_flipped_foldr[code_unfold]:
  "flip (foldr f) = flipped_foldr f"
  by (simp add: flipped_foldr_def)

lemma o_comp_infix[code_unfold]: "x \<circ> y = comp_infix x y"
  by (simp add: comp_infix_def)

code_const const (Haskell "const")
code_const flip (Haskell "flip")
code_const flipped_foldr (Haskell "foldr")
code_const "op \<circ>" (Haskell "(.)")
code_const comp_infix (Haskell infixr 9 ".")
code_const "op #" (Haskell "(:)")
code_const Cons_infix (Haskell infixr 5 ":")
(*
(*Problem: Haskell's "length" is of type "[a] -> Int", whereas
  Isabelle's "length" is of type "'a list => nat" (thus having
  arbitrary precision). Hence, no direct mapping is possible/desirable.*)
code_const length (Haskell "length")
*)

text {* NOTE: the definition of comp_infix together with o_comp_infix leads to
nontermination for all target languages that do not have a proper code_const setup
for comp_infix *}

end
