(*  Title:       Formalization of Decreasing Diagrams for Term Rewrite Systems
    Author:      Harald Zankl  <Harald.Zankl at uibk.ac.at>, 2014
    Maintainer:  Harald Zankl  <Harald.Zankl at uibk.ac.at>, 2014
*)

(*
Copyright 2014 Harald Zankl

This formalization is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.

This formalization is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with the formalization. If not, see <http://www.gnu.org/licenses/>.
*)

header "Decreasing Diagrams2"

theory Decreasing_Diagrams2 
imports 
  "$AFP/Decreasing-Diagrams/Decreasing_Diagrams" 
  Critical_Pairs
begin

(* abbreviations for hidden consts from Decrasing_Diagrams  *)
abbreviation "DD_D \<equiv> Decreasing_Diagrams.D"
abbreviation "DD_seq \<equiv> Decreasing_Diagrams.seq"

type_synonym ('f,'v) step = "('f,'v) term \<times> ('f,'v) rule \<times> pos \<times> ('f,'v) subst \<times> ('f,'v) term"
type_synonym ('f,'v) lpeak = "('f,'v) step \<times> ('f,'v) step"
type_synonym ('f,'v) join = "('f,'v) step list \<times> ('f,'v) step list"
(*a labeling is a function from rewrite steps to a set that respects an ordering*)
type_synonym ('f,'v,'a) labeling = "('f,'v) step \<Rightarrow> 'a"

definition get_source :: "('f,'v) step \<Rightarrow> ('f,'v) term"
  where "get_source s = fst s"

definition get_rule :: "('f,'v) step \<Rightarrow> ('f,'v) rule"
 where "get_rule s = fst (snd s)"

definition get_pos :: "('f,'v) step \<Rightarrow> pos"
 where "get_pos s = fst (snd (snd s))"

definition get_subst :: "('f,'v) step \<Rightarrow> ('f,'v) subst"
 where "get_subst s = fst (snd (snd (snd s)))"

definition get_target :: "('f,'v) step \<Rightarrow> ('f,'v) term"
 where "get_target s = snd (snd (snd (snd s)))"

type_synonym ('f,'v) seq = "(('f,'v) term \<times> (('f,'v) step) list)"

inductive_set seq :: "('f,'v) trs \<Rightarrow> ('f,'v) seq set" 
for R where 
  "(s,[]) \<in> seq R"
| "(s,t) \<in> rstep_r_p_s R r p \<sigma> \<Longrightarrow> (t,ts) \<in> seq R \<Longrightarrow> (s,(s,r,p,\<sigma>,t)#ts) \<in> seq R"

lemma rstep_is_seq: 
  assumes "(s,t) \<in> rstep_r_p_s R r p \<sigma>"
  shows "(s,[(s,r,p,\<sigma>,t)]) \<in> seq R" 
using assms seq.intros by blast

fun first :: "('f,'v) seq \<Rightarrow> ('f,'v) term"
  where "first (s,_) = s"

fun last :: "('f,'v) seq \<Rightarrow> ('f,'v) term"
where 
  "last (s,[]) = s"
| "last (s,step # ss) = last (get_target step,ss)"

lemma rstep_r_p_s_ctxt_closed: 
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" and "q \<in> poss u"
  shows "((ctxt_of_pos_term q u)\<langle>s\<rangle>,(ctxt_of_pos_term q u)\<langle>t\<rangle>) \<in> rstep_r_p_s R rl (q<#>p) \<sigma>"
using assms unfolding rstep_r_p_s_def'
apply auto
apply (metis hole_pos_ctxt_of_pos_term hole_pos_poss)
apply (simp add: replace_at_subt_at)
apply (metis hole_pos_ctxt_of_pos_term hole_pos_poss_conv replace_at_subt_at subterm_poss_conv)
by (metis assms(2) ctxt_compose_apply_distrib ctxt_of_pos_term_append ctxt_of_pos_term_hole_pos 
  hole_pos_ctxt_of_pos_term hole_pos_poss replace_at_subt_at)

lemma rstep_r_p_s_ctxt_closed2: 
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" 
  shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) \<sigma>"
using rstep_r_p_s_ctxt_closed[OF assms] ctxt_of_pos_term_hole_pos hole_pos_poss
by fastforce

lemma rstep_r_p_s_subst_closed:
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<tau>" 
  shows "(s\<cdot>\<sigma>,t\<cdot>\<sigma>) \<in> rstep_r_p_s R rl p (\<tau>\<circ>s \<sigma>)" 
using assms unfolding rstep_r_p_s_def' 
by (auto simp add: ctxt_of_pos_term_subst)

lemma seq_chop_first: 
  assumes "(s,(s,rl,p,\<sigma>,t) # ss) \<in> seq R"
  shows "(t,ss) \<in> seq R" and "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" 
using assms seq.cases by auto  

fun step_close:: "('f,'v) step \<Rightarrow> ('f,'v) ctxt \<Rightarrow> ('f,'v) subst \<Rightarrow> ('f,'v) step"
where 
  "step_close step C \<tau> = 
    (let (s,rl,p,\<sigma>,t) = step in
      (C\<langle>s\<cdot>\<tau>\<rangle>,rl,hole_pos C<#>p,\<sigma> \<circ>s \<tau>,C\<langle>t\<cdot>\<tau>\<rangle>))"

fun seq_close :: "('f,'v) seq \<Rightarrow> ('f,'v) ctxt \<Rightarrow> ('f,'v) subst \<Rightarrow> ('f,'v) seq"
 where "seq_close ss C \<tau> = (C\<langle>fst ss\<cdot>\<tau>\<rangle>,map (\<lambda>step. step_close step C \<tau>) (snd ss) )"

(*peaks*)
definition local_peaks :: "('f,string) trs \<Rightarrow> (('f,string) lpeak) set"
  where "local_peaks R = {((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u)) | s t u r1 r2 p1 p2 \<sigma>1 \<sigma>2. 
  ((s,t) \<in> rstep_r_p_s R r1 p1 \<sigma>1 \<and> (s,u) \<in> rstep_r_p_s R r2 p2 \<sigma>2)}"

definition critical_peaks :: "('f,string) trs \<Rightarrow> (bool \<times> ('f,string) lpeak) set"
  where "critical_peaks R = {(C = \<box>, (l \<cdot> \<sigma>, (l, r), \<epsilon>, \<sigma>, r \<cdot> \<sigma>), (l \<cdot> \<sigma>, (l', r'), hole_pos C, \<tau>, (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>)) | l r l' r' l'' C \<sigma> \<tau>.  
    (l, r) \<in> R \<and> (l', r') \<in> R \<and> l = C\<langle>l''\<rangle> \<and> is_Fun l'' \<and>
    mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>) }"

lemma critical_peak_is_local_peak:
  assumes "(b,s1,s2) \<in> critical_peaks R"
  shows "(s1,s2) \<in> local_peaks R"
using assms
unfolding critical_peaks_def local_peaks_def rstep_r_p_s_def
by auto (metis ctxt_of_pos_term_hole_pos hole_pos_subst mgu_var_disjoint_generic_sound)+

lemma critical_peaksI:
  assumes "(l, r) \<in> R" and "(l', r') \<in> R" and "l = C\<langle>l''\<rangle>"
  and "is_Fun l''" and "mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>)" and "s = r \<cdot> \<sigma>"
  and "t =  (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>" and "b = (C = \<box>)"
  shows "(b, (l \<cdot> \<sigma>, (l, r), \<epsilon>, \<sigma>, s), (l \<cdot> \<sigma>, (l', r'), hole_pos C, \<tau>, t)) \<in> critical_peaks R"
using assms unfolding critical_peaks_def by blast

lemma critical_peak_to_critical_pair:
  assumes "(b,s1,s2) \<in> critical_peaks R"
  shows "(b, get_target s1,get_target s2) \<in> critical_pairs R R"
using assms 
unfolding critical_peaks_def critical_pairs_def get_target_def 
by fastforce

definition parallel_peak :: "('f,string) trs \<Rightarrow> (('f,string) lpeak) \<Rightarrow> bool"
 where "parallel_peak R p = (
  p \<in> local_peaks R \<and>
  (let ((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u)) = p in p1 \<bottom> p2)
 )"

definition variable_peak :: "('f,string) trs \<Rightarrow> (('f,string) lpeak) \<Rightarrow> bool"
 where "variable_peak R p = (
  p \<in> local_peaks R \<and>
  (let ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) = p in
  \<exists> r. ((p1 <#> r = p2) \<and> \<not> (r \<in> poss (fst rl1) \<and> is_Fun ((fst rl1) |_ r)))
 ))"

definition function_peak :: "('f,string) trs \<Rightarrow> (('f,string) lpeak) \<Rightarrow> bool"
 where "function_peak R p = (
  p \<in> local_peaks R \<and>
  (let ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) = p in
  \<exists> r. ((p1 <#> r = p2) \<and> r \<in> poss (fst rl1) \<and> is_Fun ((fst rl1) |_ r)))
 )"

(*decreasingness*)
definition d_trs :: "('f,char list) trs \<Rightarrow> ('f,char list) lpeak \<Rightarrow> ('f,char list) seq \<Rightarrow> ('f,char list) seq \<Rightarrow> bool"
 where "d_trs R p jl jr = (p \<in> local_peaks R \<and> jl \<in> seq R \<and> jr \<in> seq R \<and> get_target (fst p) = first jl \<and> get_target (snd p) = first jr \<and> last jl = last jr)"

definition seq_concat :: "('f,'v) seq \<Rightarrow> ('f,'v) seq \<Rightarrow> ('f,'v) seq"
 where "seq_concat \<sigma>1 \<sigma>2 = (first \<sigma>1,snd \<sigma>1 @ snd \<sigma>2)"

definition LD_trs :: "('f,'v) trs \<Rightarrow> (('f,'v,'a) labeling) \<Rightarrow> 'a rel \<Rightarrow> ('f,'v) lpeak \<Rightarrow> ('f,'v) seq \<Rightarrow> ('f,'v) seq \<Rightarrow> bool"
 where "LD_trs R l r p jl jr = (\<exists> \<sigma>1 \<sigma>2 \<sigma>3 \<tau>1 \<tau>2 \<tau>3. ((map l (snd jl) = \<sigma>1@\<sigma>2@\<sigma>3) \<and> (map l (snd jr) = \<tau>1@\<tau>2@\<tau>3) \<and>
 (LD_1' r (l (fst p)) (l (snd p)) \<sigma>1 \<sigma>2 \<sigma>3) \<and>
 (LD_1' r (l (snd p)) (l (fst p)) \<tau>1 \<tau>2 \<tau>3)
))"

definition D_trs :: "('f,'v) trs \<Rightarrow> (('f,'v,'a) labeling) \<Rightarrow> 'a rel \<Rightarrow> ('f,'v) lpeak \<Rightarrow> ('f,'v) seq \<Rightarrow> ('f,'v) seq \<Rightarrow> bool"
where "D_trs R l r p jl jr = DD_D r [l (fst p)] [l (snd p)] (map l (snd jl)) (map l (snd jr))"

definition ldd :: "('f,char list) trs \<Rightarrow> ('f,char list,'a) labeling \<Rightarrow> 'a rel \<Rightarrow> ('f,char list) lpeak \<Rightarrow> bool"
where "ldd R l r p = (\<exists> jl jr. (d_trs R p jl jr \<and> D_trs R l r p jl jr))"

(*labelings*)

definition is_labeling :: "('f,char list) trs \<Rightarrow> ('f,char list,'a) labeling \<Rightarrow> 'a rel \<Rightarrow> bool"
 where "is_labeling R l r = (\<forall> s rl1 p \<sigma> t u rl2 q \<tau> v C \<rho> \<sigma>1 \<sigma>2 p1 p2 \<sigma>3. (
(*function case*)
 ((s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1 \<and> (u,v) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and>
(* (C\<langle>s\<cdot>\<rho>\<rangle>,C\<langle>t\<cdot>\<rho>\<rangle>) \<in> rstep_r_p_s R rl1 (hole_pos C<#>p1) (\<sigma>1\<circ>s \<rho>) \<and> 
 (C\<langle>u\<cdot>\<rho>\<rangle>,C\<langle>v\<cdot>\<rho>\<rangle>) \<in> rstep_r_p_s R rl2 (hole_pos C<#>p2) (\<sigma>2\<circ>s \<rho>)
 \<and>
*)
 (l(s,rl1,p1,\<sigma>1,t),l(u,rl2,p2,\<sigma>2,v)) \<in> r \<longrightarrow> 
   (l(C\<langle>s\<cdot>\<rho>\<rangle>,rl1,hole_pos C<#>p1,\<sigma>1\<circ>s \<rho>,C\<langle>t\<cdot>\<rho>\<rangle>),l(C\<langle>u\<cdot>\<rho>\<rangle>,rl2,hole_pos C<#>p2,\<sigma>2\<circ>s \<rho>,C\<langle>v\<cdot>\<rho>\<rangle>)) \<in> r)
 \<and> ((l(s,rl1,p1,\<sigma>1,t) = l(u,rl2,p2,\<sigma>2,v)) \<longrightarrow> l (step_close (s,rl1,p1,\<sigma>1,t) C \<rho>) = l (step_close (u,rl2,p2,\<sigma>2,v) C \<rho>))
\<and>
(*variable case*)
((s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1 \<and> (s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and>
 ((t = v) \<or> (t,v) \<in> rstep_r_p_s R rl2 q \<sigma>2)  \<and> (u,v) \<in> rstep_r_p_s R rl1 p1 \<sigma>3 \<longrightarrow>
 (l (s,rl1,p1,\<sigma>1,t) = l (u,rl1,p1,\<sigma>3,v)) \<and> (l (s,rl2,p2,\<sigma>2,u) = l (t,rl2,q,\<sigma>2,v))
) \<and>
(*parallel case*)
 ((s,t) \<in> rstep_r_p_s R rl1 p \<sigma> \<and> (s,u) \<in> rstep_r_p_s R rl2 q \<tau> \<and>
  (t,v) \<in> rstep_r_p_s R rl2 q \<tau> \<and> (u,v) \<in> rstep_r_p_s R rl1 p \<sigma> \<longrightarrow>
  (l (s,rl1,p,\<sigma>,t) = l (u,rl1,p,\<sigma>,v)) \<and> (l (s,rl2,q,\<tau>,u) = l (t,rl2,q,\<tau>,v)))

))"

lemma l_closed:
  fixes C \<tau>
  assumes lab: "is_labeling R l r"
  and s1: "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1"
  and ss: "ss \<in> seq R"
  and dec: "\<forall> b \<in> set (map l (snd ss)). b = l (s,rl1,p1,\<sigma>1,t)"
  shows "\<forall> b \<in> set (map l (snd (seq_close ss C \<tau>))). b = l (step_close (s,rl1,p1,\<sigma>1,t) C \<tau>)" 
using ss dec
proof (induct "snd ss" arbitrary:"ss")
 case Nil thus ?case by auto
next
 case (Cons step steps) 
 from Cons(2) obtain u where ss_dec: "ss = (u,step # steps)" by (cases ss, auto)
 with Cons(3) obtain rl2 p2 \<sigma>2 v where step_dec: "step = (u,rl2,p2,\<sigma>2,v)" and s2: "(u,v) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" 
   using seq.cases by blast
 have "l (s,rl1,p1,\<sigma>1,t) = l step" using Cons unfolding ss_dec step_dec by auto
 hence step_close: "l (step_close (s,rl1,p1,\<sigma>1,t) C \<tau>) = l (step_close step C \<tau>)" 
   using lab s1 s2 unfolding is_labeling_def step_dec by blast
 hence vs: "(v,steps) \<in> seq R" using Cons unfolding ss_dec step_dec using seq_chop_first by fast
 hence "\<forall>b\<in>set (map l (snd (v,steps))). b = l (s, rl1, p1, \<sigma>1, t)" using Cons unfolding ss_dec step_dec by auto
 hence "\<forall> b \<in> set (map l (snd (seq_close (v,steps) C \<tau>))). b = l (step_close (s,rl1,p1,\<sigma>1,t) C \<tau>)" 
   using Cons(1)[OF _ vs] Cons by auto
 thus ?case using step_close unfolding ss_dec by auto
qed

definition rlstep_r_p_s :: "('f,'v) trs \<Rightarrow> ('f,'v,'a) labeling \<Rightarrow> ('f,'v) rule \<Rightarrow> pos \<Rightarrow> ('f,'v) subst \<Rightarrow> (('f,'v) term\<times>'a\<times>('f,'v) term) set"
 where "rlstep_r_p_s R l rl p \<sigma> = {(s,a,t). (s,t) \<in> rstep_r_p_s R rl p \<sigma> \<and> a = l (s,rl,p,\<sigma>,t)}"

lemma step_super_closed1:
  fixes C \<sigma>
  shows "get_target (step_close step C \<sigma>) = C\<langle>get_target step\<cdot>\<sigma>\<rangle>" 
unfolding get_target_def Let_def by (cases step, simp)

lemma seq_super_closed1:
  fixes C \<sigma>
  assumes seq:"ss \<in> seq R" 
  shows "seq_close ss C \<sigma> \<in> seq R \<and> (first (seq_close ss C \<sigma>)) = C\<langle>first ss\<cdot>\<sigma>\<rangle> \<and> last (seq_close ss C \<sigma>) = C\<langle>last ss\<cdot>\<sigma>\<rangle>" 
using seq 
proof (induct "snd ss" arbitrary: ss)
  case Nil then obtain s where ss_dec: "ss = (s,[])" using surjective_pairing by metis 
  thus ?case using seq.intros unfolding ss_dec by auto
next
  case (Cons step ts)
  from Cons(2) obtain s where ss_dec: "ss=(s,step#ts)" by (cases ss, auto)
  with Cons(3) obtain rl p \<tau> t where step_dec: "step=(s,rl,p,\<tau>,t)"
    using seq.cases by blast
  hence step: "(s,t) \<in> rstep_r_p_s R rl p \<tau>" and  i:"(t,ts) \<in> seq R" (is "?ts \<in> _") 
    using Cons(3) seq.cases unfolding ss_dec step_dec by auto
  from Cons(1)[OF _ i] have ih: 
    "seq_close ?ts C \<sigma> \<in> seq R" "(first (seq_close ?ts C \<sigma>)) = (C\<langle>first ?ts\<cdot>\<sigma>\<rangle>)" "last (seq_close ?ts C \<sigma>) = C\<langle>last ?ts\<cdot>\<sigma>\<rangle>" 
    by auto
  have C_step: "(C\<langle>s\<cdot>\<sigma>\<rangle>,C\<langle>t\<cdot>\<sigma>\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) (\<tau>\<circ>s \<sigma>)" 
    using rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step]] by auto 
  hence C_seq: "(seq_close ss C \<sigma>) \<in> seq R" 
    using ih(1) seq.intros(2)[OF C_step] unfolding ss_dec step_dec seq_close.simps by auto
  hence C_first: "first (seq_close ss C \<sigma>) = C\<langle>first ss\<cdot>\<sigma>\<rangle>" unfolding ss_dec step_dec by auto
  have "last (seq_close ?ts C \<sigma>) = C\<langle>last ?ts\<cdot>\<sigma>\<rangle>" using ih by auto
  hence C_last: "last (seq_close ss C \<sigma>) = C\<langle>last ss\<cdot>\<sigma>\<rangle>"
    unfolding ss_dec seq_close.simps last.simps fst_conv snd_conv get_target_def  
      list.map last.simps get_target_def step_dec step_close.simps by auto
  show ?case using C_seq C_first C_last by auto
qed

lemma seq_super_closed2:
  fixes C \<sigma>
  assumes step': "(s',t') \<in> rstep_r_p_s R rl' p' \<rho>"
  and lab: "is_labeling R l r"
  and seq: "ss \<in> seq R"
  and dec:"\<forall> b \<in> set (map l (snd ss)). (b, l (s',rl',p',\<rho>,t')) \<in> r"
  shows "(\<forall> b \<in> set (map l (snd (seq_close ss C \<sigma>))). (b,l (step_close (s',rl',p',\<rho>,t') C \<sigma>)) \<in> r)" 
using seq dec 
proof (induct "snd ss" arbitrary: ss)
  case Nil then obtain s where ss_dec: "ss = (s,[])" using surjective_pairing by metis 
  thus ?case using seq.intros unfolding ss_dec by auto
next
  case (Cons step ts)
  from Cons(2) obtain s where ss_dec: "ss=(s,step#ts)" by (cases ss, auto)
  with Cons(3) obtain rl p \<tau> t where step_dec: "step=(s,rl,p,\<tau>,t)"
    using seq.cases by blast
  hence step: "(s,t) \<in> rstep_r_p_s R rl p \<tau>" and i:"(t,ts) \<in> seq R" (is "?ts \<in> _") 
    using Cons(3) seq.cases unfolding ss_dec step_dec by auto
  have j:"\<forall>b \<in> set (map l (snd ?ts)). (b, l (s',rl',p',\<rho>,t')) \<in> r" 
    using Cons(4) unfolding ss_dec snd_conv list.map by auto
  from Cons(1)[OF _ i j] have
    ih: "(\<forall> b \<in> set (map l (snd (seq_close ?ts C \<sigma>))). (b,l (C\<langle>s'\<cdot>\<sigma>\<rangle>,rl',hole_pos C<#>p',\<rho>\<circ>s \<sigma>,C\<langle>t'\<cdot>\<sigma>\<rangle>)) \<in> r)" by auto
  have C_step: "(C\<langle>s\<cdot>\<sigma>\<rangle>,C\<langle>t\<cdot>\<sigma>\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) (\<tau>\<circ>s \<sigma>)" 
    using rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step]] by auto 
  have C_label: "(l (s,rl,p,\<tau>,t),l (s',rl',p',\<rho>,t')) \<in> r" using Cons(4) unfolding ss_dec step_dec snd_conv list.map by auto
  have C_label2: "(l (C\<langle>s\<cdot>\<sigma>\<rangle>,rl,hole_pos C<#>p,\<tau>\<circ>s\<sigma>,C\<langle>t\<cdot>\<sigma>\<rangle>),l (C\<langle>s'\<cdot>\<sigma>\<rangle>,rl',hole_pos C<#>p',\<rho>\<circ>s \<sigma>,C\<langle>t'\<cdot>\<sigma>\<rangle>)) \<in> r" 
    using step step' C_step rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step']] C_label lab dec assms(1) 
    unfolding is_labeling_def by blast
  hence C_labels: "(\<forall>b\<in>set (map l (snd (seq_close ss C \<sigma>))). (b, l (C\<langle>s' \<cdot> \<sigma>\<rangle>, rl', hole_pos C <#> p', \<rho> \<circ>s \<sigma>, C\<langle>t' \<cdot> \<sigma>\<rangle>)) \<in> r)" 
    using ih(1) unfolding ss_dec step_dec seq_close.simps snd_conv step_close.simps list.map Let_def by auto  
  show ?case using C_labels by auto
qed

lemma seq_super_closed3:
  fixes C \<sigma>
  assumes step': "(s',t') \<in> rstep_r_p_s R rl' p' \<rho>"
  and step'': "(s'',t'') \<in> rstep_r_p_s R rl'' p'' \<rho>'"
  and lab: "is_labeling R l r"
  and seq: "ss \<in> seq R"
  and dec:"\<forall> b \<in> set (map l (snd ss)). (b, l (s',rl',p',\<rho>,t')) \<in> r \<or> (b, l (s'',rl'',p'',\<rho>',t'')) \<in> r"
  shows "(\<forall> b \<in> set (map l (snd (seq_close ss C \<sigma>))). ((b,l (step_close (s',rl',p',\<rho>,t') C \<sigma>)) \<in> r) \<or> 
           (b,l (step_close (s'',rl'',p'',\<rho>',t'') C \<sigma>)) \<in> r)"
using seq dec 
proof (induct "snd ss" arbitrary: ss)
  case Nil then obtain s where ss_dec: "ss = (s,[])" using surjective_pairing by metis 
  thus ?case using seq.intros unfolding ss_dec by auto
next
  case (Cons step ts)
  from Cons(2) obtain s where ss_dec: "ss = (s, step # ts)" by (cases ss, auto)
  with Cons(3) obtain rl p \<tau> t where step_dec: "step = (s, rl, p, \<tau>, t)"
    using seq.cases by blast
  hence step: "(s,t) \<in> rstep_r_p_s R rl p \<tau>"  and i:"(t,ts) \<in> seq R" (is "?ts \<in> _") 
    using Cons(3) seq.cases unfolding step_dec ss_dec by auto
  have j:"\<forall>b \<in> set (map l (snd ?ts)). (b, l (s',rl',p',\<rho>,t')) \<in> r \<or> (b, l (s'',rl'',p'',\<rho>',t'')) \<in> r" 
    using Cons(4) unfolding ss_dec snd_conv list.map by auto 
  from Cons(1)[OF _ i j] have
    ih: "(\<forall> b \<in> set (map l (snd (seq_close ?ts C \<sigma>))). ((b,l (step_close (s',rl',p',\<rho>,t') C \<sigma>)) \<in> r) \<or> 
           (b,l (step_close (s'',rl'',p'',\<rho>',t'') C \<sigma>)) \<in> r)" (is "\<forall> b \<in> _. (b,?a') \<in> r \<or> (b,?a'') \<in> r") 
    by auto
  have C_step: "(C\<langle>s\<cdot>\<sigma>\<rangle>,C\<langle>t\<cdot>\<sigma>\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) (\<tau>\<circ>s \<sigma>)" 
    using rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step]] by auto 
  have C_label: "(l (s,rl,p,\<tau>,t),l (s',rl',p',\<rho>,t')) \<in> r \<or> (l (s,rl,p,\<tau>,t),l (s'',rl'',p'',\<rho>',t'')) \<in> r" 
    using Cons(4) unfolding ss_dec step_dec snd_conv list.map by auto
  have C_label2: "(l (C\<langle>s\<cdot>\<sigma>\<rangle>,rl,hole_pos C<#>p,\<tau>\<circ>s\<sigma>,C\<langle>t\<cdot>\<sigma>\<rangle>),?a') \<in> r \<or>
                  (l (C\<langle>s\<cdot>\<sigma>\<rangle>,rl,hole_pos C<#>p,\<tau>\<circ>s\<sigma>,C\<langle>t\<cdot>\<sigma>\<rangle>),?a'') \<in> r" (is "(?a,_) \<in> r \<or> _") 
  proof (cases "(l (s,rl,p,\<tau>,t),l (s',rl',p',\<rho>,t')) \<in> r")
    case True 
    thus ?thesis using step step' C_step rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step']] lab dec
      unfolding is_labeling_def step_close.simps Let_def by fast 
  next
    case False 
    hence l: "(l (s,rl,p,\<tau>,t),l (s'',rl'',p'',\<rho>',t'')) \<in> r" using C_label by auto
    thus ?thesis using step step'' C_step rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step''],of "C" "\<sigma>"] l
      lab dec unfolding is_labeling_def step_close.simps Let_def by fast
  qed
  hence C_labels: "(\<forall>b\<in>set (map l (snd (seq_close ss C \<sigma>))). ((b,?a') \<in> r \<or> (b,?a'') \<in> r))" 
    using ih(1) unfolding ss_dec step_dec seq_close.simps snd_conv step_close.simps list.map Let_def by auto  
  show ?case using C_labels by auto
qed

lemma mirror: 
  assumes "p \<in> local_peaks R" 
  shows "(snd p,fst p) \<in> local_peaks R" 
using assms unfolding local_peaks_def by auto

lemma local_peaks_helper:
  assumes pos: "p1 \<le> p2" 
  and p: "((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u)) \<in> local_peaks R" (is "?p \<in> _")
  shows "variable_peak R ?p \<or> function_peak R ?p" 
proof -
  from pos obtain r where "(p1 <#> r) = p2" using prefix_pos_diff by fast
  thus ?thesis using p unfolding variable_peak_def function_peak_def by auto
qed

lemma local_peaks_cases:
  assumes p: "p \<in> local_peaks R"
  shows "parallel_peak R p \<or> variable_peak R p \<or> variable_peak R (snd p, fst p) \<or> function_peak R p \<or> function_peak R (snd p, fst p)" 
using assms 
proof -
  from assms obtain  s t u r1 r2 p1 p2 \<sigma>1 \<sigma>2 where "(s,t) \<in> rstep_r_p_s R r1 p1 \<sigma>1 \<and> (s,u) \<in> rstep_r_p_s R r2 p2 \<sigma>2" 
    and p_dec: "p = ((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u))" unfolding local_peaks_def by fast
  thus ?thesis 
  proof (cases "p1 \<bottom> p2")
    case True 
    hence "parallel_peak R p" 
      using p unfolding parallel_peak_def p_dec Let_def by auto
    thus ?thesis by auto
  next
    case False note F = this
    thus ?thesis 
    proof (cases "p1 \<le> p2") 
      case True thus ?thesis using local_peaks_helper p p_dec by fast 
    next
      case False 
      hence "p2 \<le> p1" using F parallel_pos by auto 
      thus ?thesis using local_peaks_helper mirror p p_dec by (metis fst_conv snd_conv)
    qed
  qed  
qed

lemma seq_concat_seq_close: 
  fixes C \<tau>
  assumes "ss = seq_concat ts us"
  shows "seq_close ss C \<tau> = seq_concat (seq_close ts C \<tau>) (seq_close us C \<tau>)" 
proof -
 from assms obtain s sq t tq u uq where ss_dec: "ss = (s,sq)" 
   and ts_dec: "ts = (t,tq)" and us_dec: "us = (u,uq)" using surjective_pairing by metis
 show ?thesis using assms unfolding ss_dec ts_dec us_dec seq_close.simps seq_concat_def by auto 
qed

lemma seq_concat_labels:
  assumes "ss = seq_concat ts us"
  shows "map l (snd ss) = (map l (snd ts)) @ (map l (snd us))" 
proof -
  from assms obtain s sq t tq u uq where ss_dec: "ss = (s,sq)" 
    and ts_dec: "ts = (t,tq)" and us_dec: "us = (u,uq)" using surjective_pairing by metis
  show ?thesis using assms unfolding ss_dec ts_dec us_dec seq_close.simps seq_concat_def by auto 
qed

lemma seq_dec:
  assumes "ss \<in> seq R"
  and "map l (snd ss) = l1@l2"
  shows "\<exists> L1 L2. ss = seq_concat L1 L2 \<and> L1 \<in> seq R \<and> L2 \<in> seq R \<and> map l (snd L1) = l1 \<and> (map l (snd L2) = l2)" 
  using assms 
proof (induct l1 arbitrary:ss)
  case Nil 
  hence L1: "(first ss,[]) \<in> seq R" (is "?L1 \<in> _") using seq.intros by auto
  hence ss_dec: "ss = seq_concat ?L1 ss" unfolding seq_concat_def first.simps snd_conv 
    apply auto by (metis first.simps surjective_pairing)
  thus ?case using Nil L1 using ss_dec by (metis append_Nil list.map(1) snd_conv) 
next
  case (Cons a ls)
  from Cons obtain s step ts where ss_dec: "ss = (s,step#ts)" by (cases ss, auto)
  with Cons(2) have "(s,step#ts) \<in> seq R" by auto
  then obtain rl p \<sigma> t where step_dec: "step = (s,rl,p,\<sigma>,t)" by (cases, auto) 
  hence step: "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" using Cons(2) unfolding ss_dec step_dec using seq.cases by auto
  have i1:"(t,ts) \<in> seq R" using Cons(2) seq_chop_first(1) unfolding ss_dec step_dec by fast
  have i2:"map l (snd (t,ts)) = ls@l2" using Cons(3) Cons unfolding ss_dec step_dec by auto 
  from Cons(1)[OF i1 i2] obtain L1 L2 where 
    IH: "(t, ts) = seq_concat L1 L2" "L1 \<in> seq R" "L2 \<in> seq R" "map l (snd L1) = ls" "map l (snd L2) = l2" by fast
  have t: "first L1 = t" using IH(1) unfolding seq_concat_def by auto
  have f1: "ss = seq_concat (s,step # snd L1) L2" using IH(1) unfolding ss_dec step_dec seq_concat_def by auto
  have f2: "(s,step # snd L1) \<in> seq R" unfolding step_dec using seq.intros(2)[OF step] IH(2) t 
    by (metis first.cases first.simps snd_conv step_dec)
  have f4: "map l (snd (s,step# snd L1)) = a # ls" using IH(4) Cons(3) unfolding ss_dec step_dec by auto
  show ?case using f1 f2 IH(3) f4 IH(5) by metis
qed

lemma seq_close_length:
  fixes C \<tau>
  shows "length (snd ss) = length (snd (seq_close ss C \<tau>))" 
unfolding seq_close.simps by auto

lemma LD_trs_closed: 
  fixes C \<tau> 
  assumes d: "d_trs R p jl jr" 
  and "LD_trs R l r p jl jr" 
  and lab: "is_labeling R l r"
  shows "LD_trs R l r ((step_close (fst p) C \<tau>),(step_close (snd p) C \<tau>)) (seq_close jl C \<tau>) (seq_close jr C \<tau>)" 
    (is "LD_trs _ _ _ ?P ?JL ?JR") 
proof -
  {
    fix s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u l1 l2 l3 jl
    let ?a = "l (s,rl1,p1,\<sigma>1,t)"
    let ?a' = "l (s,rl2,p2,\<sigma>2,u)"
    let ?A = "l (step_close (s, rl1, p1, \<sigma>1, t) C \<tau>)"
    let ?A' = "l (step_close (s, rl2, p2, \<sigma>2, u) C \<tau>)"
    assume
      lstep: "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" and
      lstep': "(s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" and
      jl_dec2: "map l (snd jl) = (l1@l2)@l3" and
      LD_l: "LD_1' r ?a ?a' l1 l2 l3" and
      jl:"jl \<in> seq R"
    obtain L1 L2 L3 where jl_dec: "jl = seq_concat (seq_concat L1 L2) L3"
      and  L1:"L1 \<in> seq R" and L2: "L2 \<in> seq R" and  L3: "L3 \<in> seq R"
      and  L1_l1: "map l (snd L1) = l1" and L2_l2: "map l (snd L2) = l2" and L3_l3:"map l (snd L3) = l3"
      using seq_dec[OF jl jl_dec2] by (metis seq_dec)
    have "length (snd L2) \<le> 1" using LD_l L2_l2 unfolding LD_1'_def by auto
    hence JL: "length (snd (seq_close L2 C \<tau>)) \<le> 1" using seq_close_length by auto
    from LD_l have
      LD1: "\<forall> b \<in> set (map l (snd L1)). (b, ?a) \<in> r" and
      LD2: "\<forall> b \<in> set (map l (snd L2)). b = ?a'" and
      LD3: "\<forall> b \<in> set (map l (snd L3)). ((b, ?a) \<in> r \<or> (b, ?a') \<in> r)"
      unfolding LD_1'_def ds_def L1_l1 L2_l2 L3_l3 by auto
    let ?sc = "\<lambda>L. seq_close L C \<tau>"
    have L1:"\<forall> b \<in> set (map l (snd (?sc L1))). (b, ?A) \<in> r"
      using seq_super_closed2[OF lstep lab L1] LD1 unfolding fst_conv by auto
    have  L2:"\<forall> b \<in> set (map l (snd (?sc L2))). (b = ?A')"
      using l_closed[OF lab _ L2] lstep' LD2 unfolding snd_conv by auto
    have  L3:"\<forall> b \<in> set (map l (snd (?sc L3))). ((b, ?A) \<in> r \<or> (b, ?A') \<in> r)"
      using seq_super_closed3[OF lstep lstep' lab L3] LD3 unfolding fst_conv by simp
    hence L: "LD_1' r ?A ?A' (map l (snd (?sc L1))) (map l (snd (?sc L2))) (map l (snd (?sc L3)))"
      unfolding LD_1'_def ds_def using L1 L2 L3 JL by auto
    have "seq_close jl C \<tau> = seq_concat (seq_concat (?sc L1) (?sc L2)) (?sc L3)"
      using seq_concat_seq_close jl_dec by metis
    hence  L2:"map l (snd (?sc jl)) = (map l (snd (?sc L1))) @ (map l (snd (?sc L2))) @ (map l (snd (?sc L3)))"
      using seq_concat_labels map_append List.append_assoc by metis
    with L have "\<exists> L1 L2 L3. LD_1' r ?A ?A' (map l (snd (?sc L1))) (map l (snd (?sc L2))) (map l (snd (?sc L3))) \<and>
      map l (snd (seq_close jl C \<tau>)) = (map l (snd (?sc L1))) @ (map l (snd (?sc L2))) @ (map l (snd (?sc L3)))" by auto
  } note * = this
  from assms have pl:"p \<in> local_peaks R" unfolding d_trs_def by auto
  then obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u where p: "p = ((s, rl1, p1, \<sigma>1, t),(s, rl2, p2, \<sigma>2, u))"
    unfolding local_peaks_def by blast
  have lstep: "(s, t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" using pl unfolding local_peaks_def p by auto
  have lstep': "(s, u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" using pl unfolding local_peaks_def p by auto
  let ?a = "l (s, rl1, p1, \<sigma>1, t)"
  let ?a' = "l (s, rl2, p2, \<sigma>2, u)"
  from assms obtain l1 l2 l3 where jl_dec2: "map l (snd jl) = (l1@l2)@l3" and LD_l: "LD_1' r ?a ?a' l1 l2 l3"
    unfolding LD_trs_def unfolding p by auto
  have jl:"jl \<in> seq R" using d unfolding d_trs_def by auto
  from assms obtain r1 r2 r3 where jr_dec2: "map l (snd jr) = (r1@r2)@r3" and LD_r: "LD_1' r ?a' ?a r1 r2 r3"
    unfolding LD_trs_def unfolding p by auto
  have jr:"jr \<in> seq R" using d unfolding d_trs_def by auto
  from *[OF lstep lstep' jl_dec2 LD_l jl] *[OF lstep' lstep jr_dec2 LD_r jr] show ?thesis unfolding LD_trs_def p by fastforce
qed

lemma d_trs_closed:
  fixes C \<tau>
  assumes "d_trs R p jl jr" 
  shows "d_trs R ((step_close (fst p) C \<tau>),(step_close (snd p) C \<tau>)) (seq_close jl C \<tau>) (seq_close jr C \<tau>)" (is "d_trs R ?P _ _") 
proof -
  from assms obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u where p_dec: "p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))" and "p \<in> local_peaks R" 
    unfolding d_trs_def local_peaks_def by fast
  hence l:"(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" and r:"(s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" unfolding local_peaks_def by auto
  from l have P1: "(C\<langle>s\<cdot>\<tau>\<rangle>,C\<langle>t\<cdot>\<tau>\<rangle>) \<in> rstep_r_p_s R rl1 (hole_pos C<#> p1) (\<sigma>1\<circ>s \<tau>)" 
    using rstep_r_p_s_ctxt_closed2 rstep_r_p_s_subst_closed by metis
  from r have P2: "(C\<langle>s\<cdot>\<tau>\<rangle>,C\<langle>u\<cdot>\<tau>\<rangle>) \<in> rstep_r_p_s R rl2 (hole_pos C<#> p2) (\<sigma>2\<circ>s \<tau>)" 
    using rstep_r_p_s_ctxt_closed2 rstep_r_p_s_subst_closed by metis
  from P1 P2 have P: "?P \<in> local_peaks R" unfolding local_peaks_def p_dec step_close.simps fst_conv snd_conv Let_def by fast
  from assms have jl:"jl \<in> seq R" and jr:"jr \<in> seq R" and l: "get_target (fst p) = first jl" 
    and r: "get_target (snd p) = first jr" and b: "last jl = last jr" unfolding d_trs_def by auto
  have JL: "seq_close jl C \<tau> \<in>  seq R" using seq_super_closed1[OF jl] by fast
  have JR: "seq_close jr C \<tau> \<in>  seq R" using seq_super_closed1[OF jr] by fast
  have L: "get_target (step_close (fst (p)) C \<tau>) = first (seq_close jl C \<tau>)" 
    using step_super_closed1 l seq_super_closed1[OF jl] by metis
  have R: "get_target (step_close (snd (p)) C \<tau>) = first (seq_close jr C \<tau>)" 
    using step_super_closed1 r seq_super_closed1[OF jr] by metis 
  have "last (seq_close jl C \<tau>) = last (seq_close jr C \<tau>)" using seq_super_closed1 jl jr b by metis 
  thus ?thesis using P JL JR L R unfolding d_trs_def by auto
qed

lemma D_trs_imp_LD_trs: 
  assumes "D_trs R l r p jl jr" 
  and t:"trans r" and i:"irrefl r" 
  shows "LD_trs R l r p jl jr" 
proof -
  from assms obtain l1 l2 l3 r1 r2 r3 where "map l (snd jl) = l1@l2@l3" and "map l (snd jr) = r1@r2@r3"
    and "LD' r (l (fst p)) (l (snd p)) l1 l2 l3 r1 r2 r3"
    unfolding D_trs_def using proposition3_4_inv[OF t i] unfolding LD'_def LD_trs_def by metis
  thus ?thesis unfolding LD'_def LD_trs_def by fast
qed

lemma LD_trs_imp_D_trs:
  assumes "LD_trs R l r p jl jr"
  and t:"trans r" and i:"irrefl r"
  shows "D_trs R l r p jl jr" 
using assms proposition3_4[OF t i] unfolding D_trs_def LD_trs_def LD'_def by auto  

lemma ldd_trs_closed:
  fixes C \<tau>
  assumes "ldd R l r p"
  and t:"trans r" and i:"irrefl r" 
  and lab: "is_labeling R l r"
  shows "ldd R l r (step_close (fst p) C \<tau>,step_close (snd p) C \<tau>)" 
proof -
  from assms(1) obtain jl jr where d: "d_trs R p jl jr" and D: "D_trs R l r p jl jr" unfolding ldd_def by auto
  from d have C_d:"d_trs R (step_close (fst p) C \<tau>,step_close (snd p) C \<tau>) (seq_close jl C \<tau>) (seq_close jr C \<tau>)" 
    (is "d_trs R ?p ?jl ?jr") using d_trs_closed by fast 
  from D have "LD_trs R l r p jl jr" using D_trs_imp_LD_trs[OF _ t i] by fast
  hence "LD_trs R l r ?p ?jl ?jr" using LD_trs_closed[OF d] t i lab by fast
  hence C_D: "D_trs R l r ?p ?jl ?jr" using LD_trs_imp_D_trs[OF _ t i] by fast
  show ?thesis using C_d C_D unfolding ldd_def by auto 
qed

(*from ARS to TRS*)
abbreviation label_trs :: "('f,'v) trs \<Rightarrow> ('f,'v,'a) labeling \<Rightarrow> (('f,'v)term \<times> 'a \<times> ('f,'v)term) set" where
 "label_trs R l \<equiv> \<Union> (rl,p,\<sigma>) \<in> UNIV. rlstep_r_p_s R l rl p \<sigma>"

lemma parallel_D: 
  assumes l: "is_labeling R l r" and t: "trans r"  and i: "irrefl r"
  and "(s,t) \<in> rstep_r_p_s R rl1 p1 s1" 
  and "(s,u) \<in> rstep_r_p_s R rl2 p2 s2"
  and "(t,v) \<in> rstep_r_p_s R rl2 p2 s2"
  and "(u,v) \<in> rstep_r_p_s R rl1 p1 s1"
  shows "D_trs R l r ((s,rl1,p1,s1,t),(s,rl2,p2,s2,u)) (t,[(t,rl2,p2,s2,v)]) (u,[(u,rl1,p1,s1,v)])" (is "D_trs R l r ?p ?jl ?jr") 
proof -
  from l have "l (t, rl2, p2, s2, v) = l (s, rl2, p2, s2, u)" using assms unfolding is_labeling_def by metis
  hence ld1: "(LD_1' r (l (fst ?p)) (l (snd ?p)) (map l []) (map l (snd ?jl)) (map l []))" (is "LD_1' r ?\<beta> ?\<alpha> ?\<sigma>1 ?\<sigma>2 ?\<sigma>3") 
    unfolding LD_1'_def by auto
  from l have " l (u, rl1, p1, s1, v) = l (s, rl1, p1, s1, t)" using assms unfolding is_labeling_def by metis
  hence ld2: "(LD_1' r (l (snd ?p)) (l (fst ?p)) (map l []) (map l (snd ?jr)) (map l []))" (is "LD_1' r _ _ ?\<tau>1 ?\<tau>2 ?\<tau>3") 
    unfolding LD_1'_def by auto
  have LD': "LD' r ?\<beta> ?\<alpha> ?\<sigma>1 ?\<sigma>2 ?\<sigma>3 ?\<tau>1 ?\<tau>2 ?\<tau>3" using ld1 ld2 unfolding LD'_def by auto
  have cc1: "seq_concat (seq_concat (t,[]) ?jl) (v,[]) = ?jl" unfolding seq_concat_def by auto
  have cc2: "seq_concat (seq_concat (u,[]) ?jr) (v,[]) = ?jr" unfolding seq_concat_def by auto
  from ld1 ld2 cc1 cc2 have ld: "LD_trs R l r ?p ?jl ?jr" 
    unfolding LD_trs_def fst_conv snd_conv by (metis append_Nil append_Nil2 list.map(1))
  thus ?thesis using proposition3_4[OF t i LD'] unfolding D_trs_def by auto
qed

lemma parallel: 
  assumes "parallel_peak R p"
  and l: "is_labeling R l r" and t: "trans r" and i: "irrefl r"
  shows "ldd R l r p" 
proof -
  from assms(1) obtain r1 p1 s1 t s r2 p2 s2 u where p: "p \<in> local_peaks R" 
    and p_dec: "p = ((s,r1,p1,s1,t),(s,r2,p2,s2,u))" and b:"p1 \<bottom> p2" 
    unfolding parallel_peak_def local_peaks_def Let_def by fast
  hence s1: "(s,t) \<in> rstep_r_p_s R r1 p1 s1" and s2: "(s,u) \<in> rstep_r_p_s R r2 p2 s2" unfolding local_peaks_def by auto
  obtain v where j1: "(t,v) \<in> rstep_r_p_s R r2 p2 s2" and j2:"(u,v) \<in> rstep_r_p_s R r1 p1 s1" 
    using parallel_steps[OF _ _ b] s1 s2 surj_pair by metis
  hence jl:"(t,[(t,r2,p2,s2,v)]) \<in> seq R" (is "?jl \<in> _") 
    and jr:"(u,[(u,r1,p1,s1,v)]) \<in> seq R" (is "?jr \<in> _") using rstep_is_seq[OF j1] rstep_is_seq[OF j2] by auto
  hence "last ?jl = last ?jr" unfolding last.simps get_target_def by auto
  hence d: "d_trs R p ?jl ?jr" using p jl jr unfolding d_trs_def get_target_def p_dec by auto
  have D: "D_trs R l r p ?jl ?jr" using parallel_D[OF l t i s1 s2 j1 j2] p_dec by auto
  from d D show ?thesis unfolding ldd_def by auto
qed

lemma trivial_peak_ldd:
  assumes eq: "t1 = t2"
  and t:"trans r" and i:"irrefl r"
  and "((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2)) \<in> local_peaks R"
  shows "ldd R lab r ((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2))" (is "ldd _ _ _ ?p")
unfolding ldd_def
proof (intro exI conjI)
  show "D_trs R lab r ?p (t1, []) (t2, [])"
    apply (rule LD_trs_imp_D_trs) unfolding LD_trs_def LD_1'_def using i t by auto
  show "d_trs R ?p (t1, []) (t2, [])"
    using assms seq.intros(1) unfolding get_target_def d_trs_def by auto
qed

lemma ancestor_peak_decompose: 
  assumes "variable_peak R p" 
  shows "\<exists> s t u l1 r1 p1 \<sigma>1 l2 r2 p2 \<sigma>2 q q1 q2 w. (
   p = ((s,(l1,r1),p1,\<sigma>1,t),(s,(l2,r2),p2,\<sigma>2,u)) \<and>
   (s,t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1 \<and>
   (s,u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2 \<and>
   p1 <#> q = p2 \<and>
   (q1<#> q2) = q \<and>
   q1 \<in> poss l1 \<and>
   l1 |_ q1 = Var w)" 
proof -
  from assms obtain s rl1 p1 s1 t rl2 p2 s2 u q where p: "p \<in> local_peaks R" and p_dec: "p = ((s,rl1,p1,s1,t),(s,rl2,p2,s2,u))"
    and q:"p1 <#> q = p2" and px: "\<not> (q \<in> poss (fst rl1) \<and> is_Fun ((fst rl1) |_ q))" 
    unfolding variable_peak_def local_peaks_def Let_def  by blast 
  then obtain l1 r1 l2 r2 where rl1: "(l1,r1) = rl1" and rl2: "(l2,r2) = rl2" using surjective_pairing by metis
  hence s1: "(s,t) \<in> rstep_r_p_s R (l1,r1) p1 s1" and s2:"(s,u) \<in> rstep_r_p_s R (l2,r2) p2 s2" 
    using p p_dec unfolding local_peaks_def by auto
  have \<sigma>1: "s |_ p1 = l1 \<cdot> s1" using s1 unfolding rl1[symmetric] rstep_r_p_s_def' by auto
  have p2: "p2 \<in> poss s" using s2 unfolding rl2[symmetric] rstep_r_p_s_def' by auto
  from px have py: "\<not>(q \<in> poss l1 \<and> is_Fun (l1 |_ q))" using rl1 by auto 
  from py obtain q1 q2 w where k: "q = q1<#>q2" and q1:"q1 \<in> poss l1" and w:"l1 |_ q1 =  Var w" 
    using \<sigma>1 p2 pos_into_subst[of l1 s1 "s |_ p1" q] poss_append_poss q[symmetric] by auto 
  thus ?thesis using s1 s2 q p_dec rl1 rl2 by fast
qed

(* close local peaks decreasingly *)
lemma ancestor_steps_lin:
  assumes xy:"(x, y) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1"
  and xz:"(x, z) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2"
  and "p1 <#> v = p2"
  and "(v1 <#> v2) = v" and "v1 \<in> poss l1" and "l1 |_ v1 = Var w"
  and lin: "linear_trs R"
  defines "\<tau> \<equiv> (\<lambda>y. if y = w then replace_at (\<sigma>1 y) v2 (r2 \<cdot> \<sigma>2) else \<sigma>1 y)"
  shows "(z, (ctxt_of_pos_term p1 x)\<langle>r1 \<cdot> \<tau>\<rangle>) \<in> (rstep_r_p_s R (l1, r1) p1 \<tau>)" (is "?g1")
    "(y = (ctxt_of_pos_term p1 x)\<langle>r1 \<cdot> \<tau>\<rangle>)  \<or> 
     (\<exists>q. (q \<in> poss r1 \<and>  (r1 |_ q = Var w) \<and> (y, (ctxt_of_pos_term p1 x)\<langle>r1 \<cdot> \<tau>\<rangle>) \<in> (rstep_r_p_s R (l2,r2) (p1<#>q<#>v2) \<sigma>2)))" (is "?g2")
proof -
  from xy[unfolded rstep_r_p_s_def'] 
  have p1: "p1 \<in> poss x" and lr1: "(l1,r1) \<in> R" and \<sigma>1: "x |_ p1 = l1 \<cdot> \<sigma>1"
    and y: "y = replace_at x p1 (r1 \<cdot> \<sigma>1)" by auto
  from xz[unfolded rstep_r_p_s_def'] 
  have p2: "p2 \<in> poss x" and lr2: "(l2,r2) \<in> R" and \<sigma>2: "x |_ p2 = l2 \<cdot> \<sigma>2"
    and z: "z = replace_at x p2 (r2 \<cdot> \<sigma>2)" by auto
  from assms(3) have v:"p2 = p1 <#> v" using less_eq_pos_def by auto
  hence l2v:"(l1 \<cdot> \<sigma>1) |_ v = l2 \<cdot> \<sigma>2" using \<sigma>1 \<sigma>2 p1 p2 by auto
  from v z \<sigma>1 have zp1:"z = replace_at x p1 (replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2))" 
    using ctxt_of_pos_term_append[OF p1] ctxt_compose_apply_distrib by force
  from assms have  v12: "v = v1 <#> v2" by auto
  from assms have  w: "l1 |_ v1 = Var w" by auto
  from assms p2 have ths: "v \<in> poss (x |_ p1)" by auto
  have v1: "v1 \<in> poss l1" using assms by auto
  from assms have  wv2: "\<sigma>1 w |_ v2 = l2 \<cdot> \<sigma>2"
    by (metis \<sigma>1 l2v p2 poss_append_poss subst_apply_term.simps(1) subt_at_append subt_at_subst v1)
  have v1l1\<sigma>1:"v1 \<in> poss (l1 \<cdot> \<sigma>1)" using v1 by simp
  let ?cr = "replace_at x p1 (r1 \<cdot> \<tau>)"
  from lin lr1 have linl1:"linear_term l1" unfolding linear_trs_def by fast
  from linear_term_replace_in_subst[OF linl1 v1 w, of \<sigma>1 \<tau>, simplified]
  have "l1 \<cdot> \<tau> = replace_at (l1 \<cdot> \<sigma>1) v1 (\<tau> w)" using \<tau>_def by auto
  also have "... = replace_at (l1 \<cdot> \<sigma>1) v1 (replace_at (\<sigma>1 w) v2 (r2 \<cdot> \<sigma>2))" using \<tau>_def by auto
  also have "... = replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2)" 
    using v12 v1 w
      ctxt_of_pos_term_append[OF v1l1\<sigma>1] 
      ctxt_compose_apply_distrib by auto
  finally have "l1 \<cdot> \<tau> = replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2)" .
  with zp1 have zp2: "z = replace_at x p1 (l1 \<cdot> \<tau>)" by auto
  hence zcr:"(z, ?cr) \<in> rstep R" using subset_rstep lr1 by auto
  have p1z: "p1 \<in> poss z" using zp1 by (metis less_eq_simps(1) p2 replace_at_below_poss v z)
  have k: "ctxt_of_pos_term p1 x = ctxt_of_pos_term p1 z" 
    using ctxt_of_pos_term_replace_at_below[OF p1] zp2 by auto
  have zcr: "(z,?cr) \<in> rstep_r_p_s R (l1,r1) p1 \<tau>" 
    unfolding rstep_r_p_s_def Let_def
    using p1z lr1 zp2 k by force
  thus "?g1" by auto
  (*second goal*)
  have g2a: "(w \<notin> vars_term r1) \<longrightarrow> (y = ?cr)" 
  proof
    assume A: "w \<notin> vars_term r1" 
    show "y = ?cr" 
    proof -
      from A have "r1 \<cdot> \<tau> = r1 \<cdot> \<sigma>1" unfolding \<tau>_def by (induct r1, auto)
      hence "?cr = y" using y by auto
      thus ?thesis by auto
    qed
  qed       
 (*third goal*)
  have g2b: "w \<in> vars_term r1 \<longrightarrow> (\<exists>q. (q \<in> poss r1 \<and>  (r1 |_ q = Var w) \<and> 
    (y, ?cr) \<in> (rstep_r_p_s R (l2,r2) (p1<#>q<#>v2) \<sigma>2)))" (is "_ \<longrightarrow> ?s")
  proof
    assume W: "w \<in> vars_term r1" 
    show "?s" 
    proof -
      from W obtain q where q: "q \<in> poss r1" and qw: "r1 |_ q = Var w" using supteq_imp_subt_at[OF supteq_Var[OF W]] by auto
      from lin lr1 have linr1:"linear_term r1" unfolding linear_trs_def by fast
      have qr1\<sigma>1:"q \<in> poss (r1 \<cdot> \<sigma>1)" using q by simp
      from linear_term_replace_in_subst[OF linr1 q qw, of \<sigma>1 \<tau>, simplified]
      have "r1 \<cdot> \<tau> = replace_at (r1 \<cdot> \<sigma>1) q (\<tau> w)" using \<tau>_def by auto
      also have "... = replace_at (r1 \<cdot> \<sigma>1) q (replace_at (\<sigma>1 w) v2 (r2 \<cdot> \<sigma>2))" using \<tau>_def by auto
      also have "... = replace_at (r1 \<cdot> \<sigma>1) (q <#> v2) (r2 \<cdot> \<sigma>2)" 
        using q qw ctxt_of_pos_term_append[OF qr1\<sigma>1] ctxt_compose_apply_distrib by auto
      finally have s3:"r1 \<cdot> \<tau> = replace_at (r1 \<cdot> \<sigma>1) (q <#> v2) (r2 \<cdot> \<sigma>2)" .
      from wv2 qw have "(r1 \<cdot> \<sigma>1) |_ (q <#> v2) = l2 \<cdot> \<sigma>2" 
        using q subt_at_append[OF qr1\<sigma>1, of v2] by auto
      hence ths: "r1 \<cdot> \<sigma>1 = (ctxt_of_pos_term (q <#> v2) (r1 \<cdot> \<sigma>1))\<langle>l2 \<cdot> \<sigma>2\<rangle>"
        using qr1\<sigma>1 ctxt_supt_id \<sigma>1 p2 poss_append_poss q qw subt_at_subst v v1 v12 w using y by metis
      have k: "(q<#>v2) \<in> poss (r1 \<cdot> \<sigma>1)" using q 
        by (metis v12 w v1 \<sigma>1 p2 poss_append_poss qr1\<sigma>1 qw subst_apply_term.simps(1) subt_at_subst v)
      from ths k lr2 have key: "(r1 \<cdot> \<sigma>1, r1 \<cdot> \<tau>) \<in> rstep_r_p_s R (l2,r2) (q<#>v2) \<sigma>2" 
        unfolding s3 rstep_r_p_s_def Let_def by auto 
      hence r: "((ctxt_of_pos_term p1 x)\<langle>r1\<cdot>\<sigma>1\<rangle>,(ctxt_of_pos_term p1 x)\<langle>r1 \<cdot> \<tau>\<rangle>) \<in> rstep_r_p_s R (l2,r2) (p1<#>q<#>v2) \<sigma>2"
        using rstep_r_p_s_ctxt_closed p1 by auto
      show ?thesis using q qw r unfolding y by auto
    qed
  qed
  show "?g2" using g2a g2b by auto
qed

lemma ancestor:
  assumes lp:"variable_peak R p"
  and lin:"linear_trs R"
  shows "\<exists> jl jr. d_trs R p jl jr" 
proof -
  obtain s l1 r1 p1 s1 t l2 r2 p2 s2 u q q1 q2 w where p_dec: "p = ((s,(l1,r1),p1,s1,t),(s,(l2,r2),p2,s2,u))"
    and r:"p1 <#> q = p2" and k: "q = q1<#>q2" and q1:"q1 \<in> poss l1" and w:"l1 |_ q1 =  Var w"
    using ancestor_peak_decompose[OF lp]  by fast 
  have p: "p \<in> local_peaks R" using lp unfolding variable_peak_def by auto
  have s1: "(s,t) \<in> rstep_r_p_s R (l1,r1) p1 s1" and s2:"(s,u) \<in> rstep_r_p_s R (l2,r2) p2 s2" 
    using p p_dec unfolding local_peaks_def by auto
  from ancestor_steps_lin[OF s1 s2 r k[symmetric] q1 w lin]  obtain v s3 q where 
    j1: "(t = v) \<or> ((t,v) \<in> rstep_r_p_s R (l2,r2) q s2)" (is "_ \<or> (?jl \<in> _)")
    and j2:"(u,v) \<in> (rstep_r_p_s R (l1,r1) p1 s3)" (is "?jr \<in> _")  by auto
  have jl:"(u,[(u,(l1,r1),p1,s3,v)]) \<in> seq R" (is "?jl \<in> _") using rstep_is_seq[OF j2] by auto 
  have jr: "((t,[(t,(l2,r2),q,s2,v)]) \<in> seq R) \<or> (t = v)" (is "?jr \<in> _ \<or> _") using rstep_is_seq j1 by auto
  thus ?thesis 
  proof
    assume jrseq: "?jr \<in> seq R"
    hence  "last ?jl = last ?jr" by (auto simp add: get_target_def)
    thus ?thesis unfolding d_trs_def get_target_def using jl jrseq p p_dec by force
  next
    assume eq:"t = v"
    have jl:"(u,[(u,(l1,r1),p1,s3,v)]) \<in> seq R" (is "?jl \<in>  _") 
      and jr: "(t,[]) \<in> seq R" (is "?jr \<in>  _") using jl seq.intros by auto
    hence "last ?jl = last ?jr" using get_target_def eq by auto
    thus ?thesis unfolding d_trs_def get_target_def using jl jr p p_dec by force
  qed
qed

lemma ancestor_D1:
  assumes d: "d_trs R ((s, (l1, r1), p1, s1, t), s, (l2, r2), p2, s2, u) (t, [(t, (l2, r2), q, s2, v)]) (u, [(u, (l1, r1), p1, s3, v)])" (is "d_trs R ?p ?jl ?jr")
  and t: "trans r" and i: "irrefl r" 
  and lab: "is_labeling R l r"
  shows "D_trs R l r ((s, (l1, r1), p1, s1, t), s, (l2, r2), p2, s2, u) (t, [(t, (l2, r2), q, s2, v)]) (u, [(u, (l1, r1), p1, s3, v)])"
proof (rule LD_trs_imp_D_trs[OF _ t i])
  obtain \<sigma>2 where \<sigma>2_dec: "[l (t, (l2, r2), q, s2, v)] = [] @ \<sigma>2 @ []" by auto
  show "LD_trs R l r ?p ?jl ?jr" 
  proof -
    from d have s1: "(s,t) \<in> rstep_r_p_s R (l1,r1) p1 s1" and  s2: "(s,u) \<in> rstep_r_p_s R (l2,r2) p2 s2" 
      and s3: "(t,v) \<in> rstep_r_p_s R (l2,r2) q s2" and s4: "(u,v) \<in> rstep_r_p_s R (l1,r1) p1 s3" 
      unfolding d_trs_def local_peaks_def get_target_def using seq.cases by auto
    from lab[unfolded is_labeling_def, rule_format, of s t "(l1, r1)" p1 s1 u v "(l2, r2)" p2 s2 \<box> Var]
    have eq1: "l (t, (l2, r2), q, s2, v) = l (s, (l2, r2), p2, s2, u)" and  eq2: "l (u, (l1, r1), p1, s3, v) = l (s, (l1, r1), p1, s1, t)"
      using s1 s2 s3 s4 by fastforce+
    have "LD_1' r (l (s, (l1, r1), p1, s1, t)) (l (s, (l2, r2), p2, s2, u)) [] [l (t, (l2,r2), q, s2, v)] []" 
      using eq1 unfolding LD_1'_def ds_def by auto
    moreover have "LD_1' r (l (s, (l2, r2), p2, s2, u)) (l (s, (l1, r1), p1, s1, t)) [] [l (u, (l1,r1), p1, s3, v)] []" 
      using eq2 unfolding LD_1'_def ds_def by auto
    ultimately show ?thesis 
      unfolding LD_trs_def fst_conv snd_conv list.map
      by (metis append_Nil append_Nil2)
  qed
qed

lemma ancestor_D2:
  assumes d: "d_trs R ((s, (l1, r1), p1, s1, t), s, (l2, r2), p2, s2, u) (t, []) (u, [(u, (l1, r1), p1, s3, v)])" (is "d_trs R ?p ?jl ?jr")
  and t: "trans r" and i: "irrefl r" 
  and lab: "is_labeling R l r"
  shows "D_trs R l r ((s, (l1, r1), p1, s1, t), s, (l2, r2), p2, s2, u) (t, []) (u, [(u, (l1, r1), p1, s3, v)])" 
proof (rule LD_trs_imp_D_trs[OF _ t i])
  show "LD_trs R l r ?p ?jl ?jr" 
  proof -
    from d have s1: "(s,t) \<in> rstep_r_p_s R (l1,r1) p1 s1" and  s2: "(s,u) \<in> rstep_r_p_s R (l2,r2) p2 s2" 
      and s3: "t = v" and s4: "(u,v) \<in> rstep_r_p_s R (l1,r1) p1 s3" 
      unfolding d_trs_def local_peaks_def get_target_def apply auto using seq.cases apply auto unfolding get_target_def by auto
    have eq2:  "l (u, (l1, r1), p1, s3, v) = l (s, (l1, r1), p1, s1, t)" using d lab s1 s2 s3 s4 
      unfolding is_labeling_def apply auto by metis
    have d1: "LD_1' r (l (s, (l1, r1), p1, s1, t)) (l (s, (l2, r2), p2, s2, u)) [] [] []" 
      unfolding LD_1'_def ds_def by auto
    have d2: "LD_1' r (l (s, (l2, r2), p2, s2, u)) (l (s, (l1, r1), p1, s1, t)) [] [l (u, (l1,r1), p1, s3, v)] []" 
      using eq2 unfolding LD_1'_def ds_def by auto
    from d1 d2 show ?thesis unfolding LD_trs_def fst_conv snd_conv list.map by (metis append_Nil append_Nil2)
  qed
qed

lemma ancestor_D: 
  assumes lp:"variable_peak R p"
  and lin:"linear_trs R"
  and t: "trans r" and i: "irrefl r" and lab: "is_labeling R l r"
  shows "ldd R l r p" 
proof -
  obtain s l1 r1 p1 s1 t l2 r2 p2 s2 u q q1 q2 w where p_dec: "p = ((s,(l1,r1),p1,s1,t),(s,(l2,r2),p2,s2,u))"
    and r:"p1 <#> q = p2" and k: "q = q1<#>q2" and q1:"q1 \<in> poss l1" and w:"l1 |_ q1 =  Var w"
    using ancestor_peak_decompose[OF lp]  by fast 
  have p: "p \<in> local_peaks R" using lp unfolding variable_peak_def by auto
  have s1: "(s,t) \<in> rstep_r_p_s R (l1,r1) p1 s1" and s2:"(s,u) \<in> rstep_r_p_s R (l2,r2) p2 s2" 
    using p p_dec unfolding local_peaks_def by auto
  from ancestor_steps_lin[OF s1 s2 r k[symmetric] q1 w lin]  obtain v s3 q where 
    j1: "(t = v) \<or> ((t,v) \<in> rstep_r_p_s R (l2,r2) q s2)" (is "_ \<or> (?jr \<in> _)")
    and j2:"(u,v) \<in> (rstep_r_p_s R (l1,r1) p1 s3)" (is "?jl \<in> _")  by auto
  from j1 have jr: "(t,v) \<in> rstep_r_p_s R (l2,r2) q s2 \<or> t = v" by auto 
  have jl:"(u,[(u,(l1,r1),p1,s3,v)]) \<in> seq R" (is "?jr \<in> _") using rstep_is_seq[OF j2] by auto 
  have jr: "((t,[(t,(l2,r2),q,s2,v)]) \<in> seq R) \<or> (t = v)" (is "?jl \<in> _ \<or> _") using rstep_is_seq jr by metis 
  thus ?thesis 
  proof (cases "?jl \<in> seq R") 
    case True 
    hence  "last ?jl = last ?jr" unfolding last.simps get_target_def by auto
    hence "d_trs R p ?jl ?jr" unfolding d_trs_def get_target_def using jl True p  unfolding p_dec by auto 
    thus ?thesis using ancestor_D1[OF _ t i lab] unfolding p_dec ldd_def by fast
  next
    case False
    hence eq:"t = v" using jr by auto
    hence jl:"(u,[(u,(l1,r1),p1,s3,v)]) \<in> seq R" (is "?jr \<in>  _") 
      and jr: "(t,[]) \<in> seq R" (is "?jl \<in>  _") using jl seq.intros by auto
    hence "last ?jl = last ?jr" unfolding last.simps get_target_def eq by auto
    hence "d_trs R p ?jl ?jr" unfolding p_dec unfolding d_trs_def get_target_def using jl jr p unfolding p_dec get_target_def by auto
    thus ?thesis using ancestor_D2[OF _ t i lab] unfolding p_dec ldd_def by fast
  qed
qed

lemma critical:
  assumes "function_peak R p"
  and cp:"\<forall> (b, p) \<in> critical_peaks R. ldd R l r p"
  and trans: "trans r" and i: "irrefl r" and lab: "is_labeling R l r"
  shows "ldd R l r p" 
proof -
  from assms(1) obtain s l1 r1 p1 \<sigma>1 t l2 r2 p2 \<sigma>2 u v where p: "p \<in> local_peaks R" 
    and p_dec:"p = ((s,(l1,r1),p1,\<sigma>1,t),(s,(l2,r2),p2,\<sigma>2,u))"
    and v: "p1 <#> v = p2" and vl1: "v \<in> poss l1" and l1v: "is_Fun (l1 |_ v)" 
    unfolding function_peak_def local_peaks_def Let_def using fst_conv snd_conv surjective_pairing by auto
  have
    p1: "p1 \<in> poss s" and \<sigma>1: "s |_ p1 = l1 \<cdot> \<sigma>1" and y: "t = replace_at s p1 (r1 \<cdot> \<sigma>1)" and lr1: "(l1,r1) \<in> R" and
    p2: "p2 \<in> poss s" and \<sigma>2: "s |_ p2 = l2 \<cdot> \<sigma>2" and z: "u = replace_at s p2 (r2 \<cdot> \<sigma>2)" and lr2: "(l2,r2) \<in> R"
    using p p_dec unfolding local_peaks_def rstep_r_p_s_def' by auto
  from v have l2v: "(l1 \<cdot> \<sigma>1) |_ v = l2 \<cdot> \<sigma>2" using \<sigma>1 \<sigma>2 p1 p2 by auto
  from l2v subt_at_subst[OF vl1] have "l1 |_ v \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2" by metis
  hence l1vs1: "l1 |_ v \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2" unfolding rstep_r_p_s_def' by auto
  from mgu_var_disjoint_string_complete[OF l1vs1]
  obtain \<mu>1 \<mu>2 \<delta> where mgu: "mgu_var_disjoint_string (l1 |_ v) l2 = Some (\<mu>1,\<mu>2)" 
    and d1:"\<sigma>1 = \<mu>1 \<circ>s \<delta>" 
    and d2:"\<sigma>2 = \<mu>2 \<circ>s \<delta>" 
    and mu12:"l1 |_ v \<cdot> \<mu>1 = l2 \<cdot> \<mu>2"
    by fast
  let ?s = "l1 \<cdot> \<mu>1"
  let ?t = "r1 \<cdot> \<mu>1"
  let ?C = "ctxt_of_pos_term v l1"
  let ?u = "(?C \<cdot>c \<mu>1)\<langle>r2 \<cdot> \<mu>2\<rangle>"
  have s_step:"(s,t) \<in> rstep_r_p_s R (l1,r1) p1 \<sigma>1" using p p_dec unfolding local_peaks_def by auto
  have s_step2:"(s,u) \<in> rstep_r_p_s R (l2,r2) p2 \<sigma>2" using p p_dec unfolding local_peaks_def by auto 
  from s_step have s_dec:"(ctxt_of_pos_term p1 s)\<langle>?s \<cdot> \<delta>\<rangle> = s" unfolding rstep_r_p_s_def'
    using d1 replace_at_ident subst_compose_apply_term_distrib by auto
  from s_step have t: "(ctxt_of_pos_term p1 s)\<langle>?t\<cdot>\<delta>\<rangle> = t" unfolding rstep_r_p_s_def'
    using d1 replace_at_ident subst_compose_apply_term_distrib by auto
  from s_step s_step2 d1 d2 have u: "(ctxt_of_pos_term p1 s)\<langle>?u\<cdot>\<delta>\<rangle> = u" unfolding rstep_r_p_s_def' apply auto
    by (metis (full_types) ctxt_compose_apply_distrib ctxt_compose_subst_compose_distrib ctxt_of_pos_term_append ctxt_of_pos_term_subst d1 d2 subst_compose_apply_term_distrib v vl1 z)
  have "v = hole_pos ?C" by (metis hole_pos_ctxt_of_pos_term vl1)
  with critical_peaksI[OF lr1 lr2 _ l1v mgu, of ?C ?t ?u "?C = \<box>"] ctxt_supt_id[OF vl1]
  have "(?C = \<box>, (?s, (l1, r1), \<epsilon>, \<mu>1, ?t) ,(?s, (l2, r2), v, \<mu>2, ?u)) \<in> critical_peaks R" 
    (is "(?b, ?cp) \<in> _") by auto
  hence ldd: "ldd R l r ?cp" using cp unfolding ldd_def by auto 
  then obtain cp where cp_dec: "cp = ?cp" by auto
  hence "(p = (step_close (fst ?cp) (ctxt_of_pos_term p1 s) \<delta>,step_close (snd ?cp) (ctxt_of_pos_term p1 s) \<delta>))" 
    using p_dec s_dec[symmetric] \<sigma>1 y p1 unfolding  v[symmetric] d1 d2 apply auto 
    by (metis subst_apply_term_ctxt_apply_distrib u)
  thus ?thesis using ldd_trs_closed trans i ldd lab unfolding cp_dec by fast 
qed

lemma left:
  assumes "variable_peak R p \<or> function_peak R p"
  and cps: "\<forall> (b, p) \<in> critical_peaks R. ldd R l r p"
  and lin:"linear_trs R"
  and t: "trans r " and i: "irrefl r" and lab: "is_labeling R l r"
  shows "ldd R l r p" 
proof (cases "variable_peak R p")
  case True thus ?thesis using ancestor lin ancestor_D[OF True lin t i lab] by fast
next
  case False hence "function_peak R p" using assms by auto
  thus ?thesis using critical t i lab cps unfolding ldd_def by fast
qed

lemma ldd_mirror: 
assumes "ldd R l r p"
 and t: "trans r" and i: "irrefl r"
shows "ldd R l r (snd p,fst p)" 
proof -
  from assms obtain jl jr where d: "d_trs R p jl jr" and D: "D_trs R l r p jl jr" unfolding ldd_def by auto 
  from d have  d: "d_trs R (snd p,fst p) jr jl" unfolding d_trs_def local_peak_def by (metis fst_conv mirror snd_conv) 
  from D have D: "D_trs R l r (snd p,fst p) jr jl" unfolding D_trs_def apply auto using mirror_D[OF t i] by auto 
  show ?thesis using d D unfolding ldd_def by fast
qed

lemma main_ldd:
  assumes cps:"\<forall> (b, p) \<in> critical_peaks R. ldd R l r p"
  and lin:"linear_trs R"
  and lab: "is_labeling R l r" 
  and t: "trans r" and i: "irrefl r"
  shows "\<forall> p \<in> local_peaks R. ldd R l r p" 
proof
  fix p 
  assume A: "p \<in> local_peaks R" 
  show "ldd R l r p" 
  proof -
    show ?thesis using local_peaks_cases[OF A]
    proof (cases "parallel_peak R p")
      case True 
      show ?thesis using parallel[OF True lab t i] unfolding ldd_def by auto
    next
      case False note po = this
      thus ?thesis using local_peaks_cases[OF A] 
      proof (cases "variable_peak R p \<or> function_peak R p")
        case True 
        show ?thesis using left[OF True cps lin t i lab] by auto
      next
        case False 
        hence F: "variable_peak R (snd p,fst p) \<or> function_peak R (snd p,fst p)" 
          using local_peaks_cases[OF A] po by auto
        hence "ldd R l r (snd p, fst p)" using left[OF F cps lin t i lab] by auto
        thus ?thesis using ldd_mirror[OF _ t i] by (metis fst_conv snd_conv surjective_pairing)
      qed
    qed
  qed
qed

lemma k: 
  assumes "(s,t) \<in> (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)^*" 
  shows "(s,t) \<in> (rstep R)^*" 
proof -
  from assms obtain n where "(s,t) \<in> (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)^^n" by auto
  hence "(s,t) \<in> (rstep R)^^n" 
  proof (induct n arbitrary: s t)
    case 0 hence "s = t" by auto 
    thus ?case by auto
  next
    case (Suc n)
    then obtain u where "(s,u) \<in> (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)^^n" 
      and step:"(u,t) \<in> (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)" by (metis relpow_Suc_E)
    hence steps: "(s,u) \<in> (rstep R)^^n" using Suc(1) by auto
    from step obtain rl p \<sigma> where step: "(u,t) \<in> rstep_r_p_s R rl p \<sigma>" by auto
    hence "(u,t) \<in> (rstep R)" by (metis rstep_r_p_s_imp_rstep)
    hence "(s,t) \<in> (rstep R)^^(Suc n)" using steps by auto
    thus ?case using step steps by auto 
  qed
  thus ?thesis using relpow_imp_rtrancl by auto
qed

lemma r: 
  assumes "(s,t) \<in> (rstep R)^*" 
  shows "(s,t) \<in> (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)^*" 
proof -
  from assms obtain n where "(s,t) \<in> rstep R^^n" by auto 
  hence "(s,t) \<in> (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)^^n" 
  proof (induct n arbitrary:s t)
    case 0 thus ?case by auto
  next
    case (Suc n) 
    then obtain u where steps: "(s,u) \<in> rstep R^^n" and step: "(u,t) \<in> rstep R" using Suc by auto 
    hence steps: "(s,u) \<in> (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)^^n" using Suc by auto
    from step obtain rl p \<sigma> where "(u,t) \<in> rstep_r_p_s R rl p \<sigma>" by (metis rstep_iff_rstep_r_p_s)
    thus ?case using steps by auto
  qed
  thus ?thesis by (metis relpow_imp_rtrancl)
qed

lemma  local_peak_decompose:
  assumes "local_peak lars P"
  shows "\<exists>s t u a b. (s,a,t) \<in> lars \<and> (s,b,u) \<in> lars \<and> P = ((s,[(a,t)]),(s,[(b,u)]))" 
proof -
  from assms obtain seql seqr where P: "P = (seql,seqr)" and fsteq:"fst seql = fst seqr" 
    and lenl:"length (snd seql) = 1" and lenr:"length (snd seqr) = 1" 
    and seql: "seql \<in> DD_seq lars" and seqr: "seqr \<in> DD_seq lars"
    using assms unfolding local_peak_def peak_def by auto
  from lenl lenr obtain sl sr where "snd seql = [sl]" and "snd seqr = [sr]"
    using length_Suc_conv[of "snd seql" 0] length_Suc_conv[of "snd seqr" 0] length_0_conv by auto
  moreover from seql obtain s where "s = fst seql" using seq.cases by auto
  ultimately obtain a t b u where 
    seql_dec: "seql = (s,[(a,t)])" and
    seqr_dec: "seqr = (s,[(b,u)])" 
    using surjective_pairing fsteq by metis
  have "(s,a,t) \<in> lars" using seql unfolding seql_dec using seq_tail1(2)[of s "(a,t)" "[]"] by auto
  moreover have "(s,b,u) \<in> lars" using seqr unfolding seqr_dec using seq_tail1(2)[of s "(b,u)" "[]"] by auto
  ultimately show ?thesis using P seql_dec seqr_dec by auto
qed

lemma local_peak_introduce:
  assumes "(s,a,t) \<in> rlstep_r_p_s R l rl1 p1 \<sigma>1" and "(s,b,u) \<in> rlstep_r_p_s R l rl2 p2 \<sigma>2"
  shows "((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) \<in> local_peaks R" and "a = l (s,rl1,p1,\<sigma>1,t)" and "b = l (s,rl2,p2,\<sigma>2,u)"
using assms unfolding local_peaks_def rlstep_r_p_s_def apply auto using surjective_pairing by metis

lemma R_step_imp_lars_step:
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<sigma>"
  shows "(s,l (s,rl,p,\<sigma>,t),t) \<in> label_trs R l"
using assms unfolding rlstep_r_p_s_def by blast

lemma R_seq_imp_lars_seq:
  assumes "(s,ss) \<in> seq R"
  shows "\<exists> as \<in> DD_seq (label_trs R l). (
    s = fst as \<and> 
    last (s,ss) = lst as \<and>
    labels as = (map l ss))" 
using assms 
proof (induct ss arbitrary:s rule:list.induct)  
  case Nil hence eq: "s = last (s,[])" by auto
  hence ss: "(s, []) \<in> DD_seq (label_trs R l)" (is "?as \<in> _") using Decreasing_Diagrams.seq.intros by auto
  hence "s = fst ?as" and "last (s,[]) = lst ?as" and "labels ?as = (map l [])" 
    using eq apply (metis fst_conv) apply (metis eq fst_conv lst_def snd_conv) unfolding labels_def by auto
  thus ?case using ss by blast
next
  case (Cons step ts)
  from Cons.prems  obtain rl p \<sigma> t where step_dec: "step = (s,rl,p,\<sigma>,t)" by (cases, auto)
  hence ts: "(t,ts) \<in> seq R" using Cons(2) unfolding step_dec using seq_chop_first(1) by metis
  hence  step: "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" using Cons(2) unfolding step_dec using seq_chop_first(2) by fast
  then obtain a where lstep:"(s,a,t) \<in> rlstep_r_p_s R l rl p \<sigma>" unfolding rlstep_r_p_s_def by auto
  hence a:"a = l step" unfolding step_dec unfolding rlstep_r_p_s_def by auto
  from lstep have "(s,a,t) \<in> label_trs R l" unfolding rlstep_r_p_s_def by blast
  hence ss1:"(s,[(a,t)]) \<in> DD_seq (label_trs R l)" (is "?ss1 \<in> _")
    unfolding rlstep_r_p_s_def using Decreasing_Diagrams.seq.intros by (metis (lifting, no_types))
  from ts obtain bs where ss2:"bs \<in> DD_seq (label_trs R l)" and "t = fst bs" and "last (t,ts) = lst bs" 
    and lab0:"labels bs = map l ts" using Cons by metis
  hence as:"(s,(a,t) # snd bs) \<in> DD_seq (label_trs R l)" (is "?as \<in> _") 
    using Decreasing_Diagrams.seq_concat[OF ss1 ss2]
    by (metis Decreasing_Diagrams.seq.intros(2) `(s, a, t) \<in> label_trs R l` fst_conv snd_conv surj_pair)
  have fst:"s = fst ?as" and lst:"last (s,step#ts) = lst ?as" 
    apply auto unfolding step_dec get_target_def lst_def
    by (metis List.last.simps `t = fst bs` `last (t, ts) = lst bs` lst_def neq_Nil_conv snd_conv)
  have lab: "labels ?as = map l (step#ts)" using lab0 a unfolding labels_def by auto 
  thus ?case using as fst lst lab by metis 
qed

lemma R_seq_imp_lars_seq2:
  assumes "ss \<in> seq R"
  shows "\<exists> as \<in> DD_seq (label_trs R l). (
   first ss = fst as \<and> 
   last ss = lst as \<and>
   labels as = (map l (snd ss)))" 
proof-
  from assms obtain t ts where ss_dec: "ss = (t,ts)" using surjective_pairing by metis 
  show ?thesis using assms R_seq_imp_lars_seq[of t ts] unfolding ss_dec first.simps last.simps snd_conv by auto
qed

lemma D_trs_imp_D:
  assumes "D_trs R l r p jl jr" 
  shows "DD_D r [l (fst p)] [l (snd p)] (map l (snd jl)) (map l (snd jr))"
using assms unfolding D_trs_def by auto

lemma ldd_imp_DD:
  assumes "lars = label_trs R l"
  and lp:"local_peak lars P" and "fst (fst P) = s" and "lst (fst P) = t" and "lst (snd P) = u"
  and P: "P = ((s,[(l (s,rl1,p1,\<sigma>1,t),t)]),s,[(l (s,rl2,p2,\<sigma>2,u),u)])"
  and ldd: "ldd R l r ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))" (is "ldd _ _ _ ?p")
  shows "\<exists>\<sigma>' \<tau>'. DD lars r (fst P,snd P,\<sigma>',\<tau>')" 
proof -
  from ldd obtain jl jr where "jl \<in> seq R" and "jr \<in> seq R" and "t = first jl" and "u = first jr" and "last jl = last jr" and 
    ld: "D_trs R l r ?p jl jr" unfolding ldd_def d_trs_def get_target_def by auto

  then obtain \<sigma>' \<tau>' where jl: "\<sigma>' \<in> DD_seq lars" and 
    jr: "\<tau>' \<in> DD_seq lars" and "t = fst \<sigma>'" and "u = fst \<tau>'" and lst: "lst \<sigma>' = lst \<tau>'" 
    and labl: "labels \<sigma>' = map l (snd jl)" and labr: "labels \<tau>' = map l (snd jr)"
    using R_seq_imp_lars_seq2 unfolding assms(1)  by metis 

  from this assms have t:"fst (fst P) = fst (snd P)" and l: "lst (fst P) = fst \<sigma>'" and r: "lst (snd P) = fst \<tau>'" and 
    b: "lst \<sigma>' = lst \<tau>'" unfolding local_peak_def peak_def by auto
  have "{fst P,snd P,\<sigma>',\<tau>'} \<subseteq> DD_seq lars" using lp jl jr unfolding local_peak_def peak_def by auto  
  hence diagram: "diagram lars (fst P,snd P,\<sigma>',\<tau>')" 
    unfolding ldd_def diagram_def peak_def Let_def using t l r b by auto 

  have pl: "labels (fst P) = [l (s, rl1, p1, \<sigma>1, t)]" unfolding P labels_def by auto
  have qr: "labels (snd P) = [l (s, rl2, p2, \<sigma>2, u)]" unfolding P labels_def by auto
  from D_trs_imp_D[OF ld] have " DD_D r [l (s, rl1, p1, \<sigma>1, t)] [l (s, rl2, p2, \<sigma>2, u)] (map l (snd jl)) (map l (snd jr))" by auto
  hence "DD_D r [l (s, rl1, p1, \<sigma>1, t)] [l (s, rl2, p2, \<sigma>2, u)] (labels \<sigma>') (labels \<tau>')" using labl labr by auto  
  hence D: "DD_D r (labels (fst P)) (labels (snd P)) (labels \<sigma>') (labels \<tau>')" unfolding D2_def using pl qr by auto
  from diagram D show ?thesis unfolding DD_def D2_def Let_def by blast
qed

lemma main:
  assumes t: "trans r" and wf: "wf r" 
  and lin: "linear_trs R"
  and lab: "is_labeling R l r" 
  and dec: "\<forall> (b, p) \<in> critical_peaks R. ldd R l r p" 
  shows "CR (rstep R)" 
proof -
  from wf t have i: "irrefl r" by (metis irrefl_def wf_not_sym)
  have lp: "\<forall>p \<in> local_peaks R. ldd R l r p" using assms main_ldd[OF dec lin lab t i] by fast
  def lars \<equiv> "label_trs R l"
  have lars_dd: "\<forall>P. (local_peak lars P \<longrightarrow> (\<exists>\<sigma>' \<tau>'. (DD lars r (fst P,snd P,\<sigma>',\<tau>'))))" 
  proof 
    fix P 
    show "local_peak lars P \<longrightarrow> (\<exists>\<sigma>' \<tau>'. (DD lars r (fst P,snd P,\<sigma>',\<tau>')))" 
    proof
      assume A: "local_peak lars P" 
      show "\<exists>\<sigma>' \<tau>'. (DD lars r (fst P,snd P,\<sigma>',\<tau>'))" 
      proof -
        from A obtain s a b t u where lstep1: "(s,a,t) \<in> lars" and lstep2:"(s,b,u) \<in> lars"  
          and P: "P = ((s,[(a,t)]),(s,[(b,u)]))" using local_peak_decompose by blast
        hence top: "fst (fst P) = s" and left: "lst (fst P) = t" and right: "lst (snd P) = u" unfolding P lst_def by auto
        from lstep1 lstep2 obtain rl1 p1 \<sigma>1 rl2 p2 \<sigma>2 where 
          step1: "(s,a,t) \<in> rlstep_r_p_s R l rl1 p1 \<sigma>1" and step2: "(s,b,u) \<in> rlstep_r_p_s R l rl2 p2 \<sigma>2" unfolding lars_def by auto
        from this have "((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) \<in> local_peaks R" (is "?p \<in> _") using local_peak_introduce(1) by metis
        hence ldd: "ldd R l r ?p" using lp by auto
        from step1 step2 have a: "a = l (s,rl1,p1,\<sigma>1,t)" and b: "b = l (s,rl2,p2,\<sigma>2,u)" unfolding rlstep_r_p_s_def by auto 
        show ?thesis using ldd_imp_DD[OF _ A top left right _ ldd] P a b lars_def by auto
      qed
    qed
  qed
  def ars \<equiv> "\<Union> (rl,p,\<sigma>) \<in> UNIV. rstep_r_p_s R rl p \<sigma>"
  have unlabel:"ars = unlabel lars" unfolding unlabel_def lars_def ars_def rlstep_r_p_s_def by auto
  have CR_ars: "CR ars" using LD_imp_D[OF t wf lars_dd] D_imp_CR using unlabel by metis
  show ?thesis 
  proof
    fix s t u 
    assume A: "(s,t) \<in> (rstep R)^*" "(s,u) \<in> (rstep R)^*" 
    show "(t,u) \<in> (rstep R)\<^sup>\<down>" using A 
    proof -
      from A have "(s,t) \<in> ars^*" "(s,u) \<in> ars^*" unfolding ars_def unlabel_def using r by auto
      from this obtain v where j1:"(t,v) \<in> ars^*" "(u,v) \<in> ars^*" using CR_ars by auto
      thus ?thesis using k j1 unfolding ars_def by auto
    qed
  qed
qed

(*concrete labelings*)
 (** rule labeling **)
fun rule_labeling :: "(('f,'v) rule \<Rightarrow> nat) \<Rightarrow> ('f,'v) step \<Rightarrow> nat"
 where "rule_labeling i (s,rl,p,\<sigma>,t) = i rl"

lemma rule_labeling_is_labeling:
  shows "is_labeling R (rule_labeling i) r" 
  unfolding is_labeling_def by auto

corollary rule_labeling_is_sound:
  defines "r \<equiv> {(n,m). n < (m::nat)}"
  assumes lin: "linear_trs R"
  and dec: "\<forall> (b, p) \<in> critical_peaks R. ldd R (rule_labeling i) r p" (is "\<forall> (b, p) \<in> _ . ldd _ ?l ?r _")
  shows "CR (rstep R)" 
proof -
  have lab: "is_labeling R (rule_labeling i) r" using rule_labeling_is_labeling  by auto
  have t: "trans r" using r_def by (metis less_trans transp_def transp_trans) 
  have wf: "wf r" using r_def by (metis wf_less) 
  show ?thesis using main[OF t wf lin lab dec] by auto
qed

 (** source labeling 
fun source_labeling :: "('f,'v) step \<Rightarrow> ('f,'v) term"
  where "source_labeling (s,rl,p,\<sigma>,t) = s"

lemma source_labeling_is_labeling:
  shows "is_labeling R source_labeling (((rstep R)^+)\<inverse>)"
  unfolding is_labeling_def apply auto
  apply (metis rsteps_subst_closed trancl_rstep_ctxt_closed) sorry

corollary source_labeling_and_SN_is_sound:
  fixes R
  defines "r \<equiv> ((rstep R)^+)\<inverse>"
  assumes lin: "linear_trs R"
  and SN: "SN (rstep R)"
  and dec: "\<forall> (b, p) \<in> critical_peaks R. ldd R source_labeling r p"
  shows "CR (rstep R)"
proof -
  have lab: "is_labeling R source_labeling r" using source_labeling_is_labeling r_def by auto
 have t: "trans r" unfolding r_def by (metis trans_converse trans_trancl)
 have "SN ((rstep R)^+)" using SN by (metis SN_trancl_SN_conv)
 hence wf: "wf r" using r_def SN_iff_wf by auto
 show ?thesis using main[OF t wf lin lab dec] by auto
qed
**)

hide_const (open) last

end

