(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Icap_Impl
imports
  Q_Restricted_Rewriting_Impl
  Icap
  Linear_Orders
  Dependency_Pair_Problem_Spec
  "$AFP/Transitive-Closure/RBT_Map_Set_Extension"
  Map_Choice
begin

subsection {*Efficient Normal Form Checking*}

lemma lhs_var_not_NF:
  assumes "l \<in> T" and "is_Var l" shows "t \<notin> NF_terms T"
proof -
  from assms obtain x where l: "l = Var x" by (cases l, auto)
  let ?\<sigma> = "subst x t"
  from assms have "l \<notin> NF_terms T" by auto
  with NF_instance[of l "?\<sigma>" "Id_on T"]
    have "l \<cdot> ?\<sigma> \<notin> NF_terms T" by auto
  thus ?thesis by (simp add: l subst_def)
qed

type_synonym ('f, 'v) term_map = "'f \<Rightarrow> ('f, 'v) term list"

definition term_map :: "('f::key, 'v) term list \<Rightarrow> ('f, 'v) term_map" where
  "term_map ts = fun_of_map (rm.\<alpha> (elem_list_to_rm (fst \<circ> the \<circ> root) ts)) []"

definition
  is_NF :: "bool \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f::key, 'v) term_map \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_NF var_cond R m = (if var_cond then (\<lambda>_. False) 
     else if R = [] then (\<lambda>_. True)
     else (\<lambda>t. \<forall>u\<in>set (supteq_list t).
       if is_Fun u then
         \<forall>l\<in>set (m (fst (the (root u)))). \<not> matches u l
       else True))"

lemma neq_root_no_match:
  assumes "is_Fun l" and "the (root l) \<noteq> the (root t)"
  shows "\<not> matches t l"
  using assms by (cases t) (force iff: matches_iff)+

lemma all_not_conv: "(\<forall>x \<in> A. \<not> P x) = (\<not> (\<exists>x \<in> A. P x))" by auto

lemma efficient_supteq_list_do_not_match:
  assumes "\<forall>l\<in>set ls. \<forall>u\<in>set (supteq_list t). the (root l) \<noteq> the (root u) \<longrightarrow> \<not> matches u l"
  shows
    "(\<forall>l\<in>set ls. \<forall>u\<in>set (supteq_list t). \<not> matches u l) \<longleftrightarrow>
     (\<forall>u\<in>set (supteq_list t). \<forall>l\<in>set (term_map ls (fst (the (root u)))).
       \<not> matches u l)"
    (is "?lhs \<longleftrightarrow> ?rhs" is "_ \<longleftrightarrow> (\<forall>u\<in>set ?subs. \<forall>l\<in>set (?ls u). \<not> matches u l)")
proof (intro iffI ballI)
  fix u l assume "\<forall>l\<in>set ls. \<forall>u\<in>set ?subs. \<not> matches u l" and "u \<in> set ?subs"
    and "l \<in> set (?ls u)"
  thus "\<not> matches u l"
    using elem_list_to_rm.rm_set_lookup[of "fst \<circ> the \<circ> root" ls "fst (the (root u))"]
    by (auto simp: o_def term_map_def rm_set_lookup_def)
next
  fix l u assume 1: "\<forall>u\<in>set ?subs. \<forall>l\<in>set (?ls u). \<not> matches u l"
    and "l \<in> set ls" and "u \<in> set ?subs"
  with assms have "the (root l) \<noteq> the (root u) \<longrightarrow> \<not> matches u l"
    and "\<forall>l\<in>set (?ls u). \<not> matches u l " by simp+
  with elem_list_to_rm.rm_set_lookup[of "fst \<circ> the \<circ> root" ls "fst (the (root u))"]
    and `l \<in> set ls`
    show "\<not> matches u l" by (auto simp: o_def term_map_def rm_set_lookup_def)
qed

lemma supteq_list_ex:
  "(\<exists>u\<in>set (supteq_list l). \<exists>\<sigma>. t \<cdot> \<sigma> = u) \<longleftrightarrow> (\<exists>\<sigma>. l \<unrhd> t \<cdot> \<sigma>)"
  unfolding supteq_list by auto

lemma is_NF_NF_trs_conv:
  "is_NF (\<exists>r\<in>set R. is_Var (fst r)) (map fst R) (term_map (map fst R)) t \<longleftrightarrow>
    t \<in> NF_trs (set R)"
  (is "is_NF ?var ?R ?map t \<longleftrightarrow> _")
proof (intro iffI allI)
  assume is_NF: "is_NF ?var ?R ?map t"
  show "t \<in> NF_trs (set R)"
  proof (cases "\<exists>r\<in>set R. is_Var (fst r)")
    case True with is_NF[unfolded is_NF_def] show ?thesis by simp
  next
    case False
    let ?ts = "map fst R"
    from False have allfun: "\<forall>s\<in>set ?ts. is_Fun s" by simp
    with neq_root_no_match
    have no_match: "\<forall>s\<in>set ?ts. \<forall>u\<in>set (supteq_list t). the (root s) \<noteq> the (root u)
        \<longrightarrow> \<not> matches u s" by blast
    note is_NF = is_NF[unfolded is_NF_def if_not_P[OF False]]
    show ?thesis
    proof (cases "(R = []) = False")
      case True
      have "\<forall>u\<in>set (supteq_list t). \<forall>l\<in>set (term_map ?ts (fst (the (root u)))). \<not> matches u l"
      proof
        fix u
        assume mem: "u \<in> set (supteq_list t)"
        thus "\<forall>l\<in>set (term_map ?ts (fst (the (root u)))). \<not> matches u l"
        proof (cases u)
          case (Var x)
          show ?thesis
          proof
            fix l
            assume "l \<in> set (term_map ?ts (fst (the (root u)))) "
            with elem_list_to_rm.rm_set_lookup[of "fst \<circ> the \<circ> root" ?ts "fst (the (root u))"]
            have "l \<in> set ?ts" by (auto simp: o_def term_map_def rm_set_lookup_def)
            hence "is_Fun l" using allfun by auto
            hence "(\<forall>\<sigma>. l \<cdot> \<sigma> \<noteq> u)" using Var by auto
            thus "\<not> matches u l" using matches_iff by blast
          qed
        next
          case (Fun f us)
          with mem is_NF[simplified, unfolded True if_False] show ?thesis by auto
        qed
      qed
      thus ?thesis
        unfolding efficient_supteq_list_do_not_match[OF no_match, symmetric]
        unfolding all_not_conv matches_iff
        unfolding supteq_list_ex by auto
    next
      case False
      hence "set R = {}" by auto
      thus ?thesis by auto
    qed
  qed
next
  assume NF_trs: "t \<in> NF_trs (set R)"
  show "is_NF (\<exists>r\<in>set R. is_Var (fst r)) (map fst R) (term_map (map fst R)) t"
  proof (cases "\<exists>r\<in>set R. is_Var (fst r)")
    case True
    then obtain l where "l \<in> lhss (set R)" and "is_Var l" by auto
    from lhs_var_not_NF[OF this] and NF_trs show ?thesis by simp
  next
    case False
    let ?ts = "map fst R"
    from False have "\<forall>s\<in>set ?ts. is_Fun s" by auto
    with neq_root_no_match
    have
      no_match: "\<forall>s\<in>set ?ts. \<forall>u\<in>set (supteq_list t). the (root s) \<noteq> the (root u)
      \<longrightarrow> \<not> matches u s" by blast
    show ?thesis
    proof (cases "(map fst R = []) = False")
      case False thus ?thesis unfolding is_NF_def by auto
    next
      case True
      from NF_trs
      show ?thesis
        unfolding is_NF_def True if_False if_not_P[OF False]
        using efficient_supteq_list_do_not_match[OF no_match, symmetric]
        unfolding all_not_conv matches_iff
        unfolding supteq_list_ex set_map by force
    qed
  qed
qed

lemma is_NF_sound[simp]:
  "is_NF (\<exists>t\<in>set T. is_Var t) T (term_map T) t \<longleftrightarrow> t \<in> NF_terms (set T)"
proof -
  let ?T = "map (\<lambda>t. (t, t)) T"
  have 1: "(\<exists>t\<in>set T. is_Var t) = (\<exists>t\<in>set ?T. is_Var (fst t))"
    by (induct T) (auto simp: o_def)
  have 2: "map fst ?T = T" by (induct T) simp_all
  have 3: "term_map T = term_map (map fst ?T)" unfolding 2 ..
  have 4: "set ?T = Id_on (set T)" by (induct T) (auto simp: o_def)
  from is_NF_NF_trs_conv[of "?T" t]
    show ?thesis unfolding 1 3 4 unfolding 2 .
qed


text {*Partially applied version of @{thm is_NF_sound}.*}
lemma is_NF_sound':
  "is_NF (\<exists>t\<in>set T. is_Var t) T (term_map T) = (\<lambda>t. t \<in> NF_terms (set T))"
  by (rule ext) (simp)

lemma NF_subset_conditions:
  fixes R :: "('f, 'v) trs" and Q :: "('f, 'v) terms"
  shows "(\<forall>l\<in>lhss R. \<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>) \<longleftrightarrow> NF_terms Q \<subseteq> NF_trs R"
proof (intro iffI allI impI subsetI)
  fix t assume 1: "\<forall>l\<in>lhss R. \<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>" and 2: "t \<in> NF_terms Q"
  show "t \<in> NF_trs R"
  proof (rule ccontr)
    assume "t \<notin> NF_trs R"
    then obtain C l and \<sigma> :: "('f, 'v) subst" where "l \<in> lhss R" and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" by blast
    with 1[THEN bspec, OF `l \<in> lhss R`]
      obtain u D and \<tau> :: "('f, 'v) subst" where "u \<in> Q" and l: "l = D\<langle>u \<cdot> \<tau>\<rangle>" by auto
    hence "(u, u) \<in> Id_on Q" by auto
    from rstep.ctxt[OF rstep.subst[OF rstep.ctxt[OF rstep.subst[OF rstep.id[OF this]]]], of C D "\<tau>" "\<sigma>",
      unfolded l[symmetric] t[symmetric]]
      show False using `t \<in> NF_terms Q` by (auto simp: NF_def)
  qed
next
  assume 1: "NF_terms Q \<subseteq> NF_trs R"
  show "\<forall>l\<in>lhss R. \<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>"
  proof
    fix l assume "l \<in> lhss R"
    then obtain r where "(l, r) \<in> R" by auto
    from rstep.id[OF this] have "l \<notin> NF_trs R" by (auto simp: NF_def)
    with 1 have "l \<notin> NF_terms Q" by blast
    then obtain C l' and \<sigma> :: "('f, 'v) subst" where "l' \<in> Q" and "l = C\<langle>l' \<cdot> \<sigma>\<rangle>" by auto
    thus "\<exists>t\<in>Q. \<exists>C \<sigma>. l = C\<langle>t \<cdot> \<sigma>\<rangle>" by auto
  qed
qed

lemma ctxt_supteq_ex: "(\<exists>C \<sigma>. t = C\<langle>u \<cdot> \<sigma>\<rangle>) \<longleftrightarrow> (\<exists>\<sigma>. t \<unrhd> u \<cdot> \<sigma>)"
  using supteq_ctxt_conv[of t] by auto

(*check "NF_terms (set Q) \<subseteq> NF_trs (set R)"*)
definition NF_subset :: "('f, 'v) term list \<Rightarrow> ('f, 'v) term list \<Rightarrow> bool" where
  "NF_subset ls Q \<equiv> \<forall>l\<in>set ls. \<exists>t\<in>set Q. \<exists>u\<in>set (supteq_list l). matches u t"

lemma NF_subset_sound[simp]:
  "NF_subset (map fst R) Q \<longleftrightarrow> NF_terms (set Q) \<subseteq> NF_trs (set R)"
proof -
  have "(\<forall>l\<in>lhss (set R). \<exists>t\<in>(set Q). \<exists>\<sigma>. l \<unrhd> t \<cdot> \<sigma>)  \<longleftrightarrow> (NF_terms (set Q) \<subseteq> NF_trs (set R))"
  using NF_subset_conditions[of "set R" "set Q"] unfolding ctxt_supteq_ex by blast
  thus ?thesis
  unfolding NF_subset_def
  unfolding matches_iff
  unfolding supteq_list_ex
  unfolding set_map
  .
qed

lemma efficient_NF_check[simp]: "lhss R \<inter> NF_terms Q = {} \<longleftrightarrow> NF_terms Q \<subseteq> NF_trs R"
  using NF_terms_subset_criterion[of "lhss R" Q]
  unfolding NF_terms_lhss .

lemma term_map_subset:
  "set (term_map ts x) \<subseteq> set ts"
  using elem_list_to_rm.rm_set_lookup[of "fst \<circ> the \<circ> root" ts x]
  by (auto simp: o_def term_map_def rm_set_lookup_def)

lemma efficient_ball:
  assumes "\<forall>t\<in>set ts. the (root t) \<noteq> the (root s) \<longrightarrow> P s t"
  shows
    "(\<forall>t\<in>set ts. P s t) \<longleftrightarrow>
     (\<forall>t\<in>set (term_map ts (fst (the (root s)))). P s t)"
  using assms
    and elem_list_to_rm.rm_set_lookup[of "fst \<circ> the \<circ> root" ts "fst (the (root s))"]
  by (auto simp: o_def term_map_def rm_set_lookup_def)

fun icap_impl ::
  "bool \<Rightarrow> (('f, string) term \<Rightarrow> bool) \<Rightarrow> ('f, string) term list
    \<Rightarrow> ('f, string) term list \<Rightarrow> (string \<Rightarrow> bool) \<Rightarrow> ('f, string) term \<Rightarrow> ('f, unit + string) term"
where
  "icap_impl nf isQnf ls S Sx (Var x) =
    (if nf \<and> Sx x then Var (Inr x) else Var (Inl ()))" 
| "icap_impl nf isQnf ls S Sx (Fun f ts) = (
    let t' = Fun f (map (icap_impl nf isQnf ls S Sx) ts) in
    if \<exists>l\<in>set ls. (case mgu_class t' l of None \<Rightarrow> False | Some \<mu> \<Rightarrow> 
       (\<forall>u\<in>set (args l). isQnf (mv_yvar u \<cdot> \<mu>)) \<and>
       (\<forall>u\<in>set S. isQnf (u \<cdot> \<mu>)))
      then Var (Inl ())
      else t')"

text{*The efficient implementation of @{const icap} is correct.*}
lemma icap_impl_sound[simp]:
  fixes Q :: "('f :: key, string) term list" and S
  defines Sx: "Sx \<equiv> \<lambda>x. (\<exists>t \<in> set S. x \<in> vars_term t)"
  shows
  "icap_impl (NF_subset (map fst R) Q)
             (is_NF (\<exists>t\<in>set Q. is_Var t) Q (term_map Q))
             (map fst R) S Sx t =
    icap (set R) (set Q) (set S) t"
proof (induct t)
  case (Var x) show ?case by (simp add: Sx)
next
  case (Fun f ts)
  hence IH: "map (icap_impl (NF_subset (map fst R) Q)
                            (is_NF (\<exists>t\<in>set Q. is_Var t) Q (term_map Q))
                            (map fst R) S Sx) ts =
    map (icap (set R) (set Q) (set S)) ts" using map_eq_conv by blast  
  {
    fix f x P
    have "(case f of None \<Rightarrow> False | Some x \<Rightarrow> P x) = (\<exists> x. f = Some x \<and> P x)"
      by (cases f, auto)
  } note id = this
  show ?case
    unfolding icap_impl.simps
    unfolding IH
    unfolding id
    unfolding is_NF_sound
    unfolding icap.simps(2)[of "set R" "set Q" "set S" f ts]
    unfolding Let_def    
      by auto
qed


definition icap_impl' :: "('f :: key, string) rules
    \<Rightarrow> ('f, string) term list \<Rightarrow> ('f,string)term list \<Rightarrow> ('f, string) term \<Rightarrow> ('f, unit + string) term" where
  "icap_impl' R Q \<equiv> (
    let ls = map fst R;
        isnf = is_NF (\<exists>t\<in>set Q. is_Var t) Q (term_map Q);
        nf = NF_subset ls Q;
        ic = icap_impl nf isnf ls
      in (\<lambda> S. let S' = map mv_xvar S;
        Sx = ceta_set_of (concat (map vars_term_list S')) in
    (\<lambda> t. ic S' Sx (mv_xvar t))))"

lemma icap_impl'_sound[simp]: "icap_impl' R Q S = icap_mv (set R) (set Q) (set S)"
  unfolding icap_impl'_def 
  unfolding Let_def
  by (rule ext, unfold icap_mv_def set_map[symmetric] icap_impl_sound[symmetric], simp) 


definition icap_impl_dpp_mv :: "('d,'f,string)dpp_ops \<Rightarrow> 'd \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> ('f,unit + string)term"
  where "icap_impl_dpp_mv I d \<equiv> 
    let QR = dpp_ops.NFQ_subset_NF_rules I d;
        qnf = dpp_ops.is_QNF I d;
        r = dpp_ops.rules I d;
        ic = icap_impl QR qnf (map fst r)    
    in (\<lambda> S. let S' = map mv_xvar S;
                 Sx = ceta_set_of (concat (map vars_term_list S'))
             in (\<lambda> t. ic S' Sx (mv_xvar t)))"

lemma icap_impl_dpp_icap_mv[simp]:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "icap_impl_dpp_mv I d s = icap_mv (set r) (set q) (set s)"
proof -
  interpret dpp_spec I by fact
  show ?thesis
    by (rule ext, 
      unfold icap_mv_def set_map[symmetric] icap_impl_sound[symmetric],
      unfold icap_impl_dpp_mv_def is_NF_sound' Let_def r q, simp)
qed    

definition icap_impl_dpp :: "('d,'f,string)dpp_ops \<Rightarrow> 'd \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> ('f,unit + string)term"
  where "icap_impl_dpp I d \<equiv> 
    let QR = dpp_ops.NFQ_subset_NF_rules I d;
        qnf = dpp_ops.is_QNF I d;
        r = dpp_ops.rules I d;
        ic = icap_impl QR qnf (map fst r)    
    in (\<lambda> S. let Sx = ceta_set_of (concat (map vars_term_list S))
             in ic S Sx)"

lemma icap_impl_dpp_icap [simp]:
  fixes I :: "('d, 'f::{show, key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
    and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "icap_impl_dpp I d s = icap (set r) (set q) (set s)"
proof -
  interpret dpp_spec I by fact
  show ?thesis
    by (rule ext, 
      unfold set_map[symmetric] icap_impl_sound[symmetric],
      unfold icap_impl_dpp_def is_NF_sound' Let_def r q, simp)
qed    

end

