theory HO
imports TermFun
begin

function log_a where
  "log_a n = (if n \<le> Suc 0 then 0 else Suc (log_a (n div 2)))"
  by pat_completeness auto

termination
proof
oops

termination by lexicographic_order

fun half where "half 0 = 0" 
  | "half (Suc 0) = 0"
  | "half (Suc (Suc n)) = Suc (half n)"

function log_b where
  "log_b n = (if n \<le> Suc 0 then 0 else Suc (log_b (half n)))"
  by pat_completeness auto

termination (* by lexicographic extension *)
oops

define_trs half.simps
thm half.trs_def
define_trs (no_sim) log_b.psimps
thm log_b.trs_def

termination
apply (trs_reduce log_b)
apply expand_trs_problem
export_termination_problem "log_b.xml"
apply (ceta_import_proof "proofs/log_b.proof.xml")
apply eval
done

function foldr :: "(nat \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat list \<Rightarrow> nat" where 
  "foldr f e [] = e"
| "foldr f e (x # xs) = f x (foldr f e xs)"
  by pat_completeness auto

termination 
proof
(* first order *)
oops

define_dps foldr.psimps
thm foldr.dps_def

termination
apply (dps_reduce foldr)
apply expand_dp_problem
export_dp_problem "foldr.xml"
apply (ceta_import_dp_proof "proofs/foldr.proof.xml")
apply eval
done

function foldl :: "(nat \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat list \<Rightarrow> nat" where 
  "foldl f e [] = e"
| "foldl f e (x # xs) = foldl f (f e x) xs"
  by pat_completeness auto

termination 
proof
oops

define_dps foldl.psimps
thm foldl.dps_def

termination 
apply (dps_reduce foldl)
apply expand_dp_problem
export_dp_problem "foldl.xml"
apply (ceta_import_dp_proof "proofs/foldl.proof.xml")
apply eval
done

fun filter (* :: "(nat \<Rightarrow> bool) \<Rightarrow> nat list \<Rightarrow> nat list" *) where 
    "filter f [] = []"
  | "filter f (x # xs) = (if f x then x # filter f xs else filter f xs)"

function sieve where 
   "sieve [] = []"
 | "sieve (x # xs) = x # sieve (filter (\<lambda> y. y mod x = 0) xs)"
  by pat_completeness auto

termination 
(*  by lexicographic_order *)
oops


define_trs (no_sim) filter.simps
thm filter.trs_def

define_dps sieve.psimps
thm sieve.dps_def

termination 
(* 
apply (dps_reduce sieve)
apply expand_dp_problem
export_dp_problem "sieve.xml"
apply (ceta_import_dp_proof "proofs/sieve.proof.xml")
apply eval
done
*)
oops

fun map where 
  "map f [] = []"
| "map f (x # xs) = f x # map f xs"

datatype 'a tree = Node 'a "('a tree list)"

lemma map_cong:  
   "\<lbrakk>\<And> x. x \<in> set xs \<Longrightarrow> f x = g x \<rbrakk> \<Longrightarrow> map f xs = map g xs" 
  by (induct xs, auto)


declare map_cong[fundef_cong] 

function mirror :: "'a tree \<Rightarrow> 'a tree" where 
  "mirror (Node a ts) = Node a (rev (map mirror ts))"
  by pat_completeness auto

termination
proof
oops

termination by lexicographic_order 
thm mirror.induct

end