(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Multiset
imports
  Term
  Term_Impl
  Multiset2
begin

text {*
  here we give some function to compute multisets instead of sets of 
  function symbols, variables, etc.
  they may be helpful for non-duplicating TRSs 
*}

fun funs_term_ms :: "('f,'v)term \<Rightarrow> 'f multiset"
  where "funs_term_ms (Var x) = {#}"
     |  "funs_term_ms (Fun f ts) = {#f#} + Union_mset (multiset_of (map funs_term_ms ts))"

fun funs_ctxt_ms :: "('f,'v)ctxt \<Rightarrow> 'f multiset"
  where "funs_ctxt_ms Hole = {#}"
     |  "funs_ctxt_ms (More f bef C aft) = {#f#} + Union_mset (multiset_of (map funs_term_ms bef)) + funs_ctxt_ms C + Union_mset (multiset_of (map funs_term_ms aft))"

lemma funs_ms_ctxt_apply: "funs_term_ms C\<langle>t\<rangle> = funs_ctxt_ms C + funs_term_ms t"
  by (induct C, auto simp: multiset_eq_iff)

fun vars_term_ms :: "('f,'v)term \<Rightarrow> 'v multiset"
  where "vars_term_ms (Var x) = {#x#}"
     |  "vars_term_ms (Fun f ts) = Union_mset (multiset_of (map vars_term_ms ts))"

lemma funs_ms_subst_apply: 
  "funs_term_ms (t \<cdot> \<sigma>) = funs_term_ms t + 
            Union_mset (image_mset (\<lambda> x. funs_term_ms (\<sigma> x)) (vars_term_ms t))" 
proof (induct t)
  case (Var x) show ?case by auto
next
  case (Fun f ts)
  let ?m = "multiset_of"
  let ?f = "funs_term_ms"
  let ?v = "vars_term_ms"
  let ?ts = "Union_mset (?m (map ?f ts))"
  let ?\<sigma>  = "Union_mset (image_mset (\<lambda>x. ?f (\<sigma> x)) (Union_mset (?m (map ?v ts))))"
  let ?ts\<sigma> = "Union_mset (?m (map (\<lambda> x. ?f (x \<cdot> \<sigma>)) ts))"
  have ind: "?ts\<sigma> = ?ts + ?\<sigma>" using Fun 
  proof (induct ts)
    case Nil show ?case by simp
  next
    case (Cons t ts)
    show ?case 
      by (simp add: Cons, unfold multiset_eq_iff, auto)
  qed
  thus ?case
    by (simp add: o_def, unfold multiset_eq_iff, auto)
qed


lemma set_of_vars_term_ms[simp]: "set_of (vars_term_ms t) = vars_term t"
  by (induct t, auto)

lemma ground_vars_term_ms_empty: "ground t = (vars_term_ms t = {#})"
  unfolding ground_vars_term_empty
  unfolding set_of_vars_term_ms[symmetric]
  by (simp del: set_of_vars_term_ms)

lemma vars_term_ms_map_term [simp]:
  "vars_term_ms (map_funs_term fg t) = vars_term_ms t"
proof (induct t)
  case (Fun f ts) 
  then show ?case by (induct ts) auto
qed simp

lemma vars_term_ms_subst[simp]: 
  "vars_term_ms (t \<cdot> \<sigma>) = Union_mset (image_mset (\<lambda> x. vars_term_ms (\<sigma> x)) (vars_term_ms t))" (is "_ = ?V t")
proof (induct t)
  case (Fun f ts)
  have IH: "map (\<lambda> t. vars_term_ms (t \<cdot> \<sigma>)) ts = map (\<lambda> t. ?V t) ts"
    by (rule map_cong[OF refl Fun])
  show ?case by (simp add: o_def IH, induct ts, auto)
qed simp


lemma vars_term_ms_subst_mono: assumes "vars_term_ms s \<le> vars_term_ms t" 
  shows "vars_term_ms (s \<cdot> \<sigma>) \<le> vars_term_ms (t \<cdot> \<sigma>)"
proof -
  from assms[unfolded mset_le_exists_conv] obtain u where t: "vars_term_ms t = vars_term_ms s + u" by auto
  show ?thesis unfolding vars_term_ms_subst unfolding t by auto
qed

lemma multiset_of_funs_term_list': "multiset_of (funs_term_list' t) = funs_term_ms t"
proof (induct t)
  case (Var x) show ?case by (simp add: funs_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: funs_term_list'.simps, insert Fun, induct ts, auto simp: funs_term_list'.simps multiset_eq_iff)
qed

lemma funs_term_ms_map_term:
  "funs_term_ms (map_funs_term fg t) = image_mset fg (funs_term_ms t)"
proof (induct t)
  case (Fun f ss)
  thus ?case by (induct ss, auto)
qed auto


lemma supteq_imp_vars_term_ms_subset: "s \<unrhd> t \<Longrightarrow> vars_term_ms t \<le> vars_term_ms s" 
proof (induct rule: supteq.induct)
  case (subt u ss t f)
  from subt(1) obtain bef aft where ss: "ss = bef @ u # aft"
    by (metis in_set_conv_decomp)
  have "vars_term_ms t \<le> vars_term_ms u" by fact
  also have "\<dots> \<le> Union_mset (multiset_of (map vars_term_ms ss))"
    unfolding ss by (simp add: ac_simps)
  also have "\<dots> = vars_term_ms (Fun f ss)" by auto
  finally show ?case by auto
qed auto

end

